/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'package:target_framework/target_framework.dart';

/// The base configuration of a [TargetServer] instance.
///
/// This class is used to configure a [TargetServer] instance. You can extend
/// this class to configure your server with a custom configuration:
///
/// ```dart
/// class Configuration extends TargetConfiguration {
///   @override
///   HttpConfiguration get http => HttpConfiguration(port: 1234);
/// }
/// ```
abstract class TargetConfiguration {
  /// The HTTP configuration for the server.
  ///
  /// See [HttpConfiguration] for details.
  HttpConfiguration get http {
    return HttpConfiguration();
  }

  /// The WebSocket configuration for the server.
  ///
  /// See [WebSocketConfiguration] for details.
  WebSocketConfiguration get ws {
    return WebSocketConfiguration();
  }

  /// The directory where static files are served from.
  ///
  /// If [null], static file serving will be disabled.
  String? get staticDirectory => 'public';

  /// The current logging level for all Target logs.
  /// Defaults to [LogLevel.info].
  LogLevel get logLevel => LogLevel.info;

  /// The default handler for 404 not found requests.
  FutureOr<Response> notFoundHandler(Request request) async {
    return Response.notFound(
      'Cannot ${request.method.value.toUpperCase()} /${request.url}',
    );
  }

  /// The default handler for 500 server errors.
  FutureOr<Response> serverErrorHandler(
    Request request,
    Object error,
    StackTrace stackTrace,
  ) async {
    return Response.serverError('${error.toString()}\n$stackTrace');
  }
}
