/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'package:target_framework/target_framework.dart';

/// A controller to handle incoming HTTP and WebSocket requests.
///
/// Each incoming request has an associated method, such as `GET` or `POST`. The
/// controller can define multiple handlers to return a response to a request.
///
/// ```
/// class UsersController extends TargetController {
///   @override
///   final String get path => '/users';
///
///   @override
///   FutureOr<Response> get(Request request) async {
///     return Response.ok('Users GET request');
///   }
/// }
/// ```
///
/// See [Request] and [Response] for more information.
abstract class TargetController implements TargetBaseController {
  List<TargetMiddleware> get middleware => [];

  /// Handles any incoming HTTP request.
  ///
  /// Note; any specific method handlers will take precedence over this handler.
  FutureOr<Response> any(Request request) async {
    throw RequestNotHandled();
  }

  /// The HTTP CONNECT method starts two-way communications with the requested
  /// resource. It can be used to open a tunnel.
  ///
  /// See https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/CONNECT
  FutureOr<Response> connect(Request request) async {
    throw RequestNotHandled();
  }

  /// The HTTP DELETE request method deletes the specified resource.
  ///
  /// See https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/DELETE
  FutureOr<Response> delete(Request request) async {
    throw RequestNotHandled();
  }

  /// The HTTP GET method requests a representation of the specified resource.
  ///
  /// Requests using GET should only be used to request data (they shouldn't
  /// include data).
  ///
  /// See https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/GET
  FutureOr<Response> get(Request request) async {
    throw RequestNotHandled();
  }

  /// The HTTP HEAD method requests the headers that would be returned if the
  /// HEAD request's URL was instead requested with the HTTP GET method.
  ///
  /// Note; If no HEAD handler is defined, but a GET handler is defined, the
  /// default behaviour of a HEAD request will be to automatically respond with
  /// the GET handler with no body length.
  ///
  /// Typically you will not need to manually implemet this handler.
  ///
  /// See https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/GET
  FutureOr<Response> head(Request request) async {
    throw RequestNotHandled();
  }

  /// The HTTP OPTIONS method requests permitted communication options for a
  /// given URL or server.
  ///
  /// See https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/OPTIONS
  FutureOr<Response> options(Request request) async {
    throw RequestNotHandled();
  }

  /// The HTTP PATCH request method applies partial modifications to a resource.
  ///
  /// See https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/PATCH
  FutureOr<Response> patch(Request request) async {
    throw RequestNotHandled();
  }

  /// The HTTP POST method sends data to the server.
  ///
  /// See https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/POST
  FutureOr<Response> post(Request request) async {
    throw RequestNotHandled();
  }

  /// The HTTP PUT request method creates a new resource or replaces a
  /// representation of the target resource with the request payload.
  ///
  /// See https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/PUT
  FutureOr<Response> put(Request request) async {
    throw RequestNotHandled();
  }

  /// Handles a successully upgraded WebSocket connection.
  ///
  /// The WebSocket connection will be closed once the [Stream] is returned.
  ///
  /// To listen to messages sent from the client, listen to the `websocket.stream`
  /// [Stream]. To send messages back to the client, emit data to the [Stream].
  ///
  /// ```
  /// Stream<Object?> ws(Request request, WebSocketHandler websocket) async* {
  ///   await for (var message in websocket.stream) {
  ///     print('Received message: $message');
  ///     yield 'Sending message to client';
  ///   }
  /// }
  /// ```
  ///
  /// See https://developer.mozilla.org/en-US/docs/Web/API/WebSocket
  Stream<Object?> ws(Request request, WebSocketHandler websocket) async* {
    throw RequestNotHandled();
  }
}
