/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'package:target_framework/target_framework.dart';

/// A controller to proxy incoming HTTP requests to another endpoint.
///
/// ```
/// class ApiProxyController extends TargetProxyController {
///   @override
///   final String get path => '/api';
///
///   @override
///   Future<Response> proxy(Request request) {
///     return Response.proxy(Uri.parse('https://api.example.com'));
///   }
/// }
/// ```
///
/// In the above example, any incoming HTTP requests to the `/api` path will be
/// proxied to the `https://api.example.com`. For example, `/api/foo/bar` will
/// be proxied to `https://api.example.com/foo/bar`.
///
/// See [Request] and [Response] for more information.
abstract class TargetProxyController implements TargetBaseController {
  List<TargetMiddleware> get middleware => [];

  /// A handler to proxy incoming HTTP requests to another endpoint.
  ///
  /// This can either return a normal [Response] (e.g. in case of an invalid
  /// HTTP method), or a special proxy response which returns a [Uri] of where
  /// to proxy the request to.
  ///
  /// See [Response.proxy].
  FutureOr<Response> proxy(Request request);

  /// The optional name of the proxy.
  ///
  /// The [name] is used in headers to identify this proxy.
  /// It should be a valid HTTP token or a hostname.
  ///
  /// Defaults to 'target_proxy'.
  String? get name => 'target_proxy';
}
