/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'dart:convert';
import 'dart:io' as io show Cookie;
import 'package:crypto/crypto.dart';
import 'package:collection/collection.dart';

import 'package:target_framework/target_framework.dart';

enum CookieSameSite {
  none,
  lax,
  strict,
}

extension CookieSameSiteExtension on CookieSameSite {
  String get value {
    switch (this) {
      case CookieSameSite.none:
        return 'None';
      case CookieSameSite.lax:
        return 'Lax';
      case CookieSameSite.strict:
        return 'Strict';
    }
  }
}

class Cookie {
  Cookie(
    this.name, {
    this.secrets,
    this.domain,
    this.path,
    this.expires,
    this.httpOnly,
    this.secure,
    this.maxAge,
    this.sameSite,
  });

  final String name;

  final List<String>? secrets;
  final String? domain;
  final DateTime? expires;
  final bool? httpOnly;
  final int? maxAge;
  final String? path;
  final bool? secure;
  final CookieSameSite? sameSite;

  bool get isSigned => secrets != null && secrets!.isNotEmpty;

  String? _unsign(String cookieValue, String secret) {
    try {
      final chunks = cookieValue.split('.');
      final hmac = Hmac(sha256, utf8.encode(secret));
      final digest = hmac.convert(utf8.encode(chunks.first));
      if (digest.toString() == chunks.last) {
        return utf8.decode(
          base64Url.decode(Uri.decodeComponent(chunks.first)),
        );
      }
    } catch (_) {
      return null;
    }
    return null;
  }

  Future<String?> parse(Request request) async {
    final cookies = request.headers['cookie'] ?? '';
    final Iterable<io.Cookie> chunks = cookies.split(';').map(
      (e) {
        return io.Cookie.fromSetCookieValue(e.trim());
      },
    );

    final cookie = chunks.firstWhereOrNull((element) {
      return element.name == name;
    });

    if (cookie == null) {
      return null;
    }

    if (secrets != null && secrets!.isNotEmpty) {
      for (final secret in secrets!) {
        final value = _unsign(cookie.value, secret);

        if (value != null) {
          return value;
        }
      }

      return null;
    }

    return cookie.value;
  }

  Future<String> serialize(
    String value, {
    String? domain,
    DateTime? expires,
    bool? httpOnly,
    int? maxAge,
    String? path,
    bool? secure,
    CookieSameSite? sameSite,
  }) async {
    io.Cookie cookie = io.Cookie(name, value);

    if (secrets != null && secrets!.isNotEmpty) {
      final encoded = Uri.encodeComponent(base64UrlEncode(utf8.encode(value)));
      final hmac = Hmac(sha256, utf8.encode(secrets![0])); // HMAC-SHA256
      final digest = hmac.convert(utf8.encode(encoded));
      cookie = io.Cookie(
        name,
        '$encoded.${digest.toString()}',
      );
    }

    cookie.domain = domain ?? this.domain;
    cookie.expires = expires ?? this.expires;
    cookie.httpOnly = httpOnly ?? this.httpOnly ?? false;
    cookie.maxAge = maxAge ?? this.maxAge;
    cookie.path = path ?? this.path;
    cookie.secure = secure ?? this.secure ?? false;
    CookieSameSite? sameSiteValue = sameSite ?? this.sameSite;

    return '${cookie.toString()}${sameSiteValue != null ? '; SameSite=${sameSiteValue.value}' : ''}';
  }
}
