import 'dart:async';
import 'dart:convert';
/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'dart:typed_data';

import 'package:target_framework/src/request.dart';
import 'package:target_framework/target_framework.dart' show ImmutableHeaders;
import 'package:shelf_multipart/form_data.dart' as multipart;
import 'package:freezed_annotation/freezed_annotation.dart';

part 'form_data.freezed.dart';

/// A [FormData] instance returned from calling [Request.formData] on an incoming
/// HTTP request.
///
/// A single form data entry can contain multiple [StringValue] or [FileValue] instances,
/// which are used to represent the form data. A single data entry value is
/// represented as a [FormDataEntryValue] instance, which allows for conditionally
/// handling form data depending on the type:
///
/// ```dart
/// Future<Response> post(Request request) async {
///   final formData = await request.formData();
///
///   final name = formData.get('name');
///
///   if (name != null) {
///     name.map(
///       file: (file) => print(file.filename),
///       string: (str) => print(str.value),
///     );
///   }
/// }
/// ```
class FormData {
  FormData([Map<String, List<FormDataEntryValue>>? params])
      : _params = {...?params};

  final Map<String, List<FormDataEntryValue>> _params;

  /// Returns whether the form data contains the given key.
  bool has(String name) {
    return _params.containsKey(name);
  }

  /// Returns all of the keys in the form data.
  Iterable<String> get keys {
    return _params.keys;
  }

  /// Returns the value of the given key.
  Iterable<List<FormDataEntryValue>> get values {
    return _params.values;
  }

  /// Returns the first [FormDataEntryValue] instance for the given key, or
  /// null if the key does not exist.
  FormDataEntryValue? get(String key) {
    return _params[key]?.first;
  }

  /// Returns a list of [FormDataEntryValue] instances for the given key.
  List<FormDataEntryValue> getAll(String key) {
    return _params[key] ?? [];
  }

  /// Gets a form data entry by name.
  List<FormDataEntryValue>? operator [](String key) {
    return _params[key];
  }
}

/// Adds an extension to a [File] class and provided utility functions
/// for managing file uploads.
extension FileValueExtension on FileValue {
  /// Returns the name of the file.
  String get name => formData.filename!;

  /// Returns the headers asscocted with the file.
  ImmutableHeaders get headers => ImmutableHeaders(formData.part.headers);

  /// Reads the file contents into a [Uint8List].
  Future<Uint8List> bytes() => formData.part.readBytes();

  /// Reads the file contents into a [String].
  Future<String> text([Encoding? encoding]) =>
      formData.part.readString(encoding);

  /// Subscribes to the file contents via a [StreamSubscription].
  StreamSubscription<List<int>> listen(
    void Function(List<int> data)? onData, {
    void Function()? onDone,
    Function? onError,
    bool? cancelOnError,
  }) {
    return formData.part.listen(
      onData,
      onDone: onDone,
      onError: onError,
      cancelOnError: cancelOnError,
    );
  }
}

extension StringValueExtension on StringValue {
  String get value => string;
}

/// Represents a form data entry.
@freezed
class FormDataEntryValue with _$FormDataEntryValue {
  /// Creates a [FormDataEntryValue] instance from a [multipart.FormData].
  const factory FormDataEntryValue.file(multipart.FormData formData) =
      FileValue;

  /// Creates a [FormDataEntryValue] instance from a [String].
  const factory FormDataEntryValue.string(String string) = StringValue;
}
