/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/// HTTP headers let the client and the server pass additional information with
/// an HTTP request or response.
class Headers {
  /// Constructs a new [Headers] instance from a [Map].
  Headers([Map<String, String>? headers])
      : _headers = {...?headers}.map(
          (key, value) => MapEntry(key.toLowerCase(), value),
        );

  /// Creates a new [Headers] instance from another [Headers] instance.
  factory Headers.fromHeaders(Headers headers) {
    return Headers(headers._headers);
  }

  /// Clones the current [Headers] instance.
  Headers clone() {
    return Headers.fromHeaders(this);
  }

  final Map<String, String> _headers;

  /// Appends or creates a new header value with a given name.
  ///
  /// If the header already exists, the value will be appended to the existing
  /// header value using a comma separator.
  void append(String name, String value) {
    final _name = name.toLowerCase();
    if (_headers.containsKey(_name)) {
      _headers[_name] = '${_headers[_name]}, $value';
    } else {
      _headers[_name] = value;
    }
  }

  /// Deletes a header with a given name.
  void delete(String name) {
    _headers.remove(name.toLowerCase());
  }

  /// Returns true if the header exists.
  bool has(String name) {
    return _headers.containsKey(name.toLowerCase());
  }

  /// Returns all header keys.
  Iterable<String> get keys {
    return _headers.keys;
  }

  /// Returns all header values.
  Iterable<String> get values {
    return _headers.values;
  }

  /// Returns a [Map] of all headers.
  Map<String, String> toMap() {
    return {..._headers};
  }

  /// Sets a new header with a name and value.
  operator []=(String name, String value) {
    _headers[name.toLowerCase()] = value;
  }

  /// Gets a header value by name.
  String? operator [](String name) {
    return _headers[name.toLowerCase()];
  }
}

/// An immutable Headers instance.
///
/// Prevents any mutable operations on the headers. Typically an alternative
/// API exists for altering headers if a [ImmutableHeaders] instance is
/// provided.
class ImmutableHeaders extends Headers {
  ImmutableHeaders([Map<String, String>? headers]) : super(headers);

  @override
  void operator []=(String name, String value) {
    throw UnsupportedError('Cannot modify immutable headers.');
  }

  @override
  void append(String name, String value) {
    throw UnsupportedError('Cannot modify immutable headers.');
  }

  @override
  void delete(String name) {
    throw UnsupportedError('Cannot modify immutable headers.');
  }
}
