/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'dart:convert';

import 'package:async/async.dart';
import 'package:http_parser/http_parser.dart';
import 'package:shelf/shelf.dart' as shelf show Request;
import 'package:target_framework/target_framework.dart';
import 'package:shelf_multipart/form_data.dart' as multipart;

import 'shelf_request_view.dart';

/// Caches a [shelf.Request] and overides the [read] method allowing for
/// multipart requests to be read across the application.
///
/// Also provides some additional functionality for reading form data.
class CachedRequest extends ShelfRequestView {
  CachedRequest(shelf.Request delegate) : super(delegate);

  /// Returns whether the request is 'application/x-www-form-urlencoded' request.
  bool get isFormUrlEncoded {
    final rawContentType = headers['Content-Type'];
    if (rawContentType == null) return false;
    final contentType = MediaType.parse(rawContentType);
    return contentType.mimeType == 'application/x-www-form-urlencoded';
  }

  /// Returns a [FormData] instance from a request.
  Future<FormData> formData() async {
    if (isFormUrlEncoded) {
      final body = await readAsString();
      final params = Uri(query: body).queryParametersAll;
      return FormData(
        {
          for (final entry in params.entries)
            entry.key: [
              for (final value in entry.value) FormDataEntryValue.string(value)
            ],
        },
      );
    }

    if (isMultipartForm) {
      final Map<String, List<FormDataEntryValue>> params = {};
      await for (final data in multipartFormData) {
        params[data.name] ??= [];
        if (data.filename == null) {
          params[data.name]!
              .add(FormDataEntryValue.string(await data.part.readString()));
        } else {
          params[data.name]!.add(FormDataEntryValue.file(data));
        }
      }
      return FormData(params);
    }

    return FormData();
  }

  @override
  CachedRequest change(
      {Map<String, Object?>? headers,
      Map<String, Object?>? context,
      String? path,
      body}) {
    return CachedRequest(super
        .change(headers: headers, context: context, path: path, body: body));
  }

  /// Returns the current [StreamSplitter] instance for this request.
  StreamSplitter<List<int>> get splitter {
    final splitter =
        delegate.context['target.splitter'] as StreamSplitter<List<int>>?;

    if (splitter == null) {
      throw Exception('No "target.splitter" context was found on the request.');
    }

    return splitter;
  }

  void set splitter(StreamSplitter<List<int>> newSplitter) {
    // First close the existing splitter.
    splitter.close();

    delegate = delegate.change(context: {
      'target.splitter': newSplitter,
    });
  }

  @override
  Stream<List<int>> read() {
    return splitter.split();
  }

  @override
  Future<String> readAsString([Encoding? encoding]) {
    encoding ??= this.encoding ?? utf8;
    return encoding.decodeStream(read());
  }
}
