/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'package:target_framework/target_framework.dart';

/// Middleware allows the intercepting of requests and responses and optionally
/// modifying the request and/or response.
///
/// Middlware can be created by extending the [Middleware] class.
///
/// ```dart
/// class UserMiddleware extends Middleware {
///   @override
///   Future<Response?> requestHandler(Request request) async {
///     // Do something with the request (e.g. get a user using headers).
///     final user = await getUser(request);
///
///     if (user == null) {
///       return Response.unauthorized();
///     }
///   }
/// }
/// ```
abstract class TargetMiddleware {
  /// Called before the request is handled by a controller.
  ///
  /// If [null] is returned, the request will move on to the next middleware (or
  /// controller handler). If a [Response] is returned, the request flow will be stopped and the
  /// response immediately returned.
  ///
  /// Typically a middleware request handler will be used to check that the
  /// incoming request is as expected (e.g. checking the request headers for a
  /// user token) and attaching data to the request (see [Reqeuest.attach]).
  FutureOr<Response?> requestHandler(Request request) async {
    return null;
  }

  /// Called after a response has been returned by a controller.
  ///
  /// The middleware response handler is typically used to return a modified
  /// version of the original response, such as modyfing the content from a proxied
  /// response.
  ///
  /// Response handlers are called in reverse order of the middleware stack. For
  /// example even though the request handler for middleware A is called before
  /// middleware B, the reponse handler will be called after middleware B.
  ///
  /// `A -> B -> Controller -> B -> A`
  FutureOr<Response> responseHandler(Request request, Response response) async {
    return response;
  }
}
