/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'package:target_framework/target_framework.dart';

/// A [CORS](https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS) middleware
/// which enables CORS requests with configurable options.
///
/// Apply this middlware to your Target HTTP configuration or controller
/// middleware chain to enable CORS requests.
///
/// ```dart
/// class UserController extends TargetController {
///   @override
///   List<TargetMiddleware> get middleware => [
///     CorsMiddleware(
///       origin: '*', // enable all origins
///       allowedHeaders: ['Content-Type', 'Authorization'],
///       credentials: true,
///     ),
///   ];
/// }
/// ```
class CorsMiddleware extends TargetMiddleware {
  CorsMiddleware({
    this.origin,
    this.configureOrigin,
    this.methods = const [
      HttpMethod.get,
      HttpMethod.head,
      HttpMethod.put,
      HttpMethod.patch,
      HttpMethod.post,
      HttpMethod.delete,
    ],
    this.credentials = false,
    this.allowedHeaders = const [],
    this.exposedHeaders = const [],
    this.maxAge,
    this.optionsSuccessStatus = 204,
  });

  /// Configures the [`Access-Control-Allow-Origin`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Origin)
  /// header.
  ///
  /// If the value is set to `*` this will allow all origins to access the
  /// resource. It is recommended to set this to a specific origin. If your
  /// application serves multiple origins, you should consider using
  /// [configureOrigin] instead.
  final String? origin;

  /// Manually configures the [`Access-Control-Allow-Origin`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Origin)
  /// header. Note this overrides any provided [origin] value.
  ///
  /// If your application serves multiple origins, you should consider using
  /// this to conditionally enable CORS for a specific origin. Returning `null`
  /// does not set a value for the header.
  ///
  /// ```dart
  /// CorsMiddleware(
  ///   configureOrigin: (request) {
  ///     final origin = request.headers['Origin'];
  ///     if (orign == 'https://example.com' || origin == 'https://example.org') {
  ///       return origin;
  ///     }
  ///   },
  /// );
  /// ```
  final String? Function(Request request)? configureOrigin;

  /// Configures the [`Access-Control-Allow-Methods`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Methods)
  /// header.
  ///
  /// By default, the middleware allows GET, HEAD, PUT, PATCH, POST, and DELETE.
  final List<HttpMethod> methods;

  /// Configures the [`Access-Control-Allow-Credentials`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Credentials)
  /// header. By default, this is set to `false` and no header value is set.
  final bool credentials;

  /// Configures the [`Access-Control-Allow-Headers`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Headers)
  /// header.
  final List<String> allowedHeaders;

  /// Configures the [`Access-Control-Expose-Headers`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Expose-Headers)
  /// header.
  final List<String> exposedHeaders;

  /// Configures the [`Access-Control-Max-Age`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Max-Age)
  /// header. By default this header value is not set.
  final int? maxAge;

  /// Provides a custom status code for the OPTIONS request. By default this is
  /// set to 204.
  ///
  /// Some legacy browsers do not support a 204 response.
  final int optionsSuccessStatus;

  void _configureOrigin(Request request, Headers headers) {
    if (configureOrigin != null) {
      final configuredOrigin = configureOrigin!(request);
      if (configuredOrigin != null) {
        headers[HttpHeaders.accessControlAllowOriginHeader] = configuredOrigin;
      }
    } else {
      if (origin == '*') {
        headers[HttpHeaders.accessControlAllowOriginHeader] = '*';
      } else if (origin != null) {
        headers[HttpHeaders.accessControlAllowOriginHeader] = origin!;
      }
    }
  }

  void _configureCredentials(Request request, Headers headers) {
    if (credentials) {
      headers[HttpHeaders.accessControlAllowCredentialsHeader] = 'true';
    }
  }

  void _configureMethods(Request request, Headers headers) {
    if (methods.isNotEmpty) {
      headers[HttpHeaders.accessControlAllowMethodsHeader] =
          methods.map((m) => m.value.toUpperCase()).join(', ');
    }
  }

  void _configureAllowedHeaders(Request request, Headers headers) {
    if (allowedHeaders.isNotEmpty) {
      headers[HttpHeaders.accessControlAllowHeadersHeader] =
          allowedHeaders.join(', ');
    } else {
      final requestHeaders =
          request.headers[HttpHeaders.accessControlRequestHeadersHeader];
      if (requestHeaders != null) {
        headers[HttpHeaders.accessControlAllowHeadersHeader] = requestHeaders;
        headers.append(
          HttpHeaders.varyHeader,
          HttpHeaders.accessControlRequestHeadersHeader,
        );
      }
    }
  }

  void _configureExposedHeaders(Request request, Headers headers) {
    if (exposedHeaders.isNotEmpty) {
      headers[HttpHeaders.accessControlExposeHeadersHeader] =
          exposedHeaders.join(', ');
    }
  }

  void _configureMaxAge(Request request, Headers headers) {
    if (maxAge != null) {
      headers[HttpHeaders.accessControlMaxAgeHeader] = maxAge!.toString();
    }
  }

  @override
  Future<Response?> requestHandler(Request request) async {
    if (request.method == HttpMethod.options) {
      Headers headers = request.headers.clone();

      _configureOrigin(request, headers);
      _configureCredentials(request, headers);
      _configureMethods(request, headers);
      _configureAllowedHeaders(request, headers);
      _configureExposedHeaders(request, headers);
      _configureMaxAge(request, headers);

      return Response(
        null,
        status: optionsSuccessStatus,
        headers: headers..[HttpHeaders.contentLengthHeader] = '0',
      );
    }
    return null;
  }

  @override
  Future<Response> responseHandler(Request request, Response response) async {
    Headers headers = response.headers.clone();

    _configureOrigin(request, headers);
    _configureCredentials(request, headers);
    _configureExposedHeaders(request, headers);

    return Response(
      response.body,
      status: response.status,
      headers: headers,
    );
  }
}
