/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'dart:convert';

import 'package:target_framework/server.dart';
import 'package:target_framework/src/utils.dart';
import 'package:target_framework/target_framework.dart';
import 'package:shelf/shelf.dart' as shelf show Request;
import 'package:shelf_router/shelf_router.dart';
import 'package:shelf_multipart/multipart.dart';
import 'package:async/async.dart';

import 'internal/cached_request.dart';

/// A private extension to access internal [Request] utilities.
extension RequestExtension on Request {
  /// Returns the underlying [shelf.Request] instance.
  shelf.Request get delegate => _delegate;

  /// Returns the underlying [StreamSplitter] instance from context.
  StreamSplitter<List<int>> get splitter => _delegate.splitter;
}

/// Creates a new [Request] from the given [shelf.Request].
Request createRequest(
  TargetServer target,
  shelf.Request shelfRequest,
) {
  return Request._(target, shelfRequest);
}

/// Represents an incoming HTTP request.
class Request {
  Request._(this.target, shelf.Request shelfRequest)
      : _delegate = CachedRequest(shelfRequest);

  /// The [TargetServer] instance handling this request.
  final TargetServer target;

  /// The underlying [CachedRequest] instance.
  CachedRequest _delegate;

  Map<RequestContextKey, Object?> get _targetContext {
    if (_delegate.context['target.context'] == null) {
      return {};
    } else {
      return Map.from(_delegate.context['target.context'] as Map);
    }
  }

  /// Attaches context to the request using a [RequestContextKey].
  ///
  /// The context can later be read using [get].
  ///
  /// ```dart
  /// final key = RequestContextKey<String>();
  /// request.set(key, 'value');
  /// final value = request.get(key);
  /// ```
  void attach<T>(RequestContextKey<T> key, T value) {
    _delegate = _delegate.change(
      context: {
        'target.context': {
          ..._targetContext,
          key: value,
        },
      },
    );
  }

  /// Gets a context value which has been attached to the request using [attach].
  ///
  /// If the [RequestContextKey] has a default value and no value has been attached,
  /// the default value will be returned.
  ///
  /// If [get] is called when there is no default value, and no value has been
  /// attached, an [Exception] will be thrown.
  T get<T>(RequestContextKey<T> key) {
    final context = _targetContext;

    if (!context.containsKey(key) && key.defaultValue != null) {
      return context[key] = key.defaultValue!(this);
    }

    final value = context[key];

    if (value is! T) {
      throw Exception(
        'No valid value matching the RequestContextKey was found.',
      );
    }

    return value;
  }

  /// The [HttpMethod] of the request.
  HttpMethod get method => stringToHttpMethod(_delegate.method);

  /// The params of the request.
  ///
  /// Assuming a path of `/users/<id>`, the params would be `{'id': '...'}`.
  Map<String, String> get params => _delegate.params;

  /// The url of the request, represented as a [Uri].
  Uri get url => _delegate.url;

  /// The headers of the request.
  ///
  /// The request headers are immutable. To modify the headers of a request (
  /// for example within middleware, see [change]).
  ImmutableHeaders get headers {
    return ImmutableHeaders(_delegate.headers);
  }

  /// Returns the body of the request as a [Stream] of bytes.
  Stream<List<int>> stream() => _delegate.read();

  /// Returns the body of the request as a [Future] of a [String].
  Future<String> text() => _delegate.readAsString();

  /// Changes the current request.
  void change({Headers? headers, List<int>? body}) {
    // If a body is passed, create a new split stream.
    if (body != null) {
      _delegate.splitter = StreamSplitter(Stream.fromIterable([body]));
    }

    _delegate = CachedRequest(
      shelf.Request(
        _delegate.method,
        _delegate.requestedUri,
        protocolVersion: _delegate.protocolVersion,
        url: _delegate.url,
        encoding: _delegate.encoding,
        headers: headers?.toMap() ?? _delegate.headers,
        context: _delegate.context,
        // Body needs to be updated to keep headers such as content-length in sync.
        body: body ?? _delegate.read(),
      ),
    );
  }

  /// Returns the body of the request as a JSON object.
  ///
  /// If the body is not a valid JSON object, an [Exception] will be thrown.
  Future<Object> json() async {
    String body = await _delegate.readAsString();
    return jsonDecode(body);
  }

  /// Returns a [FormData] instance of the request.
  Future<FormData> formData() async {
    return _delegate.formData();
  }

  /// Returns the [Multipart] headers of the request.
  Stream<Multipart> multipart() {
    if (_delegate.isMultipart) {
      return _delegate.parts;
    }

    return Stream<Multipart>.empty();
  }
}
