/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'dart:convert';
import 'dart:io';

import 'package:mime/mime.dart';
import 'package:shelf/shelf.dart' as shelf;
import 'package:target_framework/src/view.dart';
import 'package:target_framework/target_framework.dart';

/// Represents a proxy response.
///
/// This class should not be used directly. Instead use [Request.proxy] whilst
/// within a [TargetProxyController].
class ResponseProxy extends Response {
  ResponseProxy(this.uri) : super(null);

  /// The [Uri] to proxy the request to.
  final Uri uri;
}

extension ResponseExtension on Response {
  shelf.Response get delegate => _delegate;
}

class Response {
  Response(
    Object? body, {
    int status = 200,
    Headers? headers,
  }) : this._delegate = shelf.Response(
          status,
          body: body,
          headers: headers?.toMap(),
        );

  Response._(this._delegate);

  final shelf.Response _delegate;

  int get status => _delegate.statusCode;

  ImmutableHeaders get headers => ImmutableHeaders(_delegate.headers);

  Stream<List<int>> get body => _delegate.read();

  static ResponseProxy proxy(Uri uri) => ResponseProxy(uri);

  factory Response.fromShelfResponse(shelf.Response response) {
    return Response._(response);
  }

  factory Response.ok(
    Object? body, {
    Headers? headers,
    Encoding? encoding,
  }) {
    return Response._(shelf.Response.ok(
      body,
      headers: headers?.toMap(),
      encoding: encoding,
    ));
  }

  factory Response.stream(
    Stream<List<int>> stream, {
    int status = 200,
    Headers? headers,
    Encoding? encoding,
  }) {
    return Response._(shelf.Response(
      status,
      body: stream,
      headers: headers?.toMap(),
      encoding: encoding,
    ));
  }

  factory Response.json(
    Object? json, {
    int status = 200,
    Headers? headers,
    Encoding? encoding,
  }) {
    final responseHeaders = (headers?.clone() ?? Headers())
      ..['content-type'] = 'application/json; charset=utf-8';
    return Response._(shelf.Response(
      status,
      body: jsonEncode(json),
      headers: responseHeaders.toMap(),
      encoding: encoding,
    ));
  }

  factory Response.redirect(
    String url, {
    bool permanent = false,
    Headers? headers,
  }) {
    final responseHeaders = (headers?.clone() ?? Headers())..['location'] = url;

    return Response._(shelf.Response(
      permanent ? 301 : 302,
      body: null,
      headers: responseHeaders.toMap(),
    ));
  }

  factory Response.badRequest(
    Object? body, {
    Headers? headers,
  }) {
    return Response._(shelf.Response(
      400,
      body: body,
      headers: headers?.toMap(),
    ));
  }

  factory Response.unauthorized(
    Object? body, {
    Headers? headers,
  }) {
    return Response._(shelf.Response(
      401,
      body: body,
      headers: headers?.toMap(),
    ));
  }

  factory Response.forbidden(
    Object? body, {
    Headers? headers,
  }) {
    return Response._(shelf.Response(
      403,
      body: body,
      headers: headers?.toMap(),
    ));
  }

  factory Response.notFound(
    Object? body, {
    Headers? headers,
  }) {
    return Response._(shelf.Response(
      404,
      body: body,
      headers: headers?.toMap(),
    ));
  }

  factory Response.serverError(
    Object? body, {
    Headers? headers,
  }) {
    return Response._(shelf.Response(
      500,
      body: body,
      headers: headers?.toMap(),
    ));
  }

  factory Response.file(File file) {
    if (!file.existsSync()) {
      return Response.notFound('File not found.');
    }

    return Response._(shelf.Response(
      200,
      body: file.openRead(),
      headers: Headers(
        {
          'content-type':
              lookupMimeType(file.path) ?? 'application/octet-stream'
        },
      ).toMap(),
    ));
  }

  factory Response.view(String path, [Map<String, dynamic>? data]) {
    final view = View(path, data ?? {});

    return Response._(shelf.Response(
      200,
      body: view.build(),
      headers: Headers({
        'Content-Type': 'text/html',
      }).toMap(),
    ));
  }
}
