/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'dart:async';
import 'dart:io';

import 'package:target_framework/src/request.dart';
import 'package:target_framework/src/response.dart';
import 'package:target_framework/src/utils.dart';
import 'package:target_framework/target_framework.dart';
import 'package:async/async.dart';
import 'package:path/path.dart' as path;
import 'package:shelf/shelf.dart' as shelf;
import 'package:shelf/shelf_io.dart' as shelf_io;
import 'package:shelf_router/shelf_router.dart' as shelf_router;
import 'package:shelf_static/shelf_static.dart' as shelf_static;
import 'package:shelf_proxy/shelf_proxy.dart' as shelf_proxy;
import 'package:web_socket_channel/status.dart' as ws_status;

import 'internal/request_logger.dart';
import 'websocket.dart';

/// An internally used configuration instance incase a user controller
/// is not provided.
class _DefaultConfiguration extends TargetConfiguration {}

extension TargetServerExtension on TargetServer {
  /// Overrides any existing controllers on the server.
  ///
  /// Note; you must call [start] to start the server once called for these
  /// controller updates to take effect.
  void setControllers(List<TargetBaseController> controllers) {
    _controllers = controllers;
  }
}

class TargetServer {
  TargetServer({
    required TargetConfiguration? configuration,
    required List<TargetBaseController> controllers,
  })  : this._controllers = controllers,
        this.configuration = configuration ?? _DefaultConfiguration();

  /// The configuration for the server.
  ///
  /// If not specified, the default configuration will be used.
  final TargetConfiguration configuration;

  /// A list of HTTP Controllers for the server to use.
  List<TargetBaseController> _controllers;

  List<TargetBaseController> get controllers => _controllers;

  /// The current server instance, if started.
  HttpServer? server;

  /// Handles any not found requests and returns a [shelf.Response].
  ///
  /// Internally, the provided configuration not found handler is called.
  Future<shelf.Response> _notFoundHandler(shelf.Request shelfRequest) async {
    Response response = await configuration.notFoundHandler(
      createRequest(this, shelfRequest),
    );

    return response.delegate;
  }

  /// Handles any server errors and returns a [shelf.Response].
  ///
  /// Internally, the provided configuration server error handler is called.
  Future<shelf.Response> _serverErrorHandler(
    shelf.Request shelfRequest,
    Object error,
    StackTrace stackTrace,
  ) async {
    Response response = await configuration.serverErrorHandler(
      createRequest(this, shelfRequest),
      error,
      stackTrace,
    );

    return response.delegate;
  }

  /// Removes the body of a request (by changing the content-length) header.
  ///
  /// This is useful if attempting to return a HEAD request to the client.
  shelf.Response _removeResponseBody(shelf.Response response) {
    if (response.headers.containsKey(HttpHeaders.contentLengthHeader)) {
      response = response.change(
        headers: {HttpHeaders.contentLengthHeader: '0'},
      );
    }
    response = response.change(body: []);
    return response;
  }

  /// Creates a pipeline for the server.
  ///
  /// Middleware can be added to the pipeline by calling [addMiddleware].
  shelf.Pipeline _createServerPipeline() {
    shelf.Pipeline pipeline = const shelf.Pipeline();

    // Add a request logger as the very fist middlware in the pipeline.
    pipeline = pipeline.addMiddleware(logRequest());

    // Split the original shelf stream so it's available to be read more than once.
    pipeline = pipeline.addMiddleware((shelf.Handler innerHandler) {
      return (shelf.Request shelfRequest) async {
        shelfRequest = shelfRequest.change(context: {
          'target.splitter': StreamSplitter(shelfRequest.read()),
        });

        return Future.sync(() {
          return innerHandler(shelfRequest);
        });
      };
    });

    // Add the server middleware to the pipeline
    pipeline = createMiddlewarePipeline(
      pipeline,
      this,
      configuration.http.middleware,
    );

    return pipeline;
  }

  /// Creates a handler for serving static files from a directory.
  ///
  /// If [null] is returned, no static files will be served.
  shelf.Handler? _createStaticFileHandler() {
    final directory = configuration.staticDirectory;

    if (directory == null) {
      return null;
    }

    Directory staticDir = Directory(
      path.join(targetApplicationDirectory, directory),
    );

    if (staticDir.existsSync()) {
      return shelf_static.createStaticHandler(staticDir.path);
    }

    return null;
  }

  void _assertNotResponseProxy(Response response) {
    if (response is ResponseProxy) {
      throw Exception(
        'The usage of Response.proxy is only supported within a TargetProxyController',
      );
    }
  }

  /// Starts the HttpServer.
  Future<HttpServer> start() async {
    // Create a router instance with the default not found handler.
    final router = shelf_router.Router(
      notFoundHandler: _notFoundHandler,
    );

    // TODO move the controller logic elsewhere
    // Create a new route for each controller.
    for (final controller in controllers) {
      shelf.Pipeline controllerPipeline = shelf.Pipeline();

      controllerPipeline = createMiddlewarePipeline(
        controllerPipeline,
        this,
        controller.middleware,
      );

      if (controller is TargetProxyController) {
        // TODO: Add path validation (e.g. cannot be regex)
        router.mount(
          controller.path,
          controllerPipeline.addHandler((shelf.Request shelfRequest) async {
            Request request = createRequest(
              this,
              shelfRequest,
            );

            var res;
            final response = await controller.proxy(request);

            if (response is ResponseProxy) {
              final proxy = shelf_proxy.proxyHandler(
                response.uri,
                proxyName: controller.name,
              );
              res = await proxy(request.delegate);
            } else {
              res = response.delegate;
            }

            request.splitter.close();
            return res;
          }),
        );
        break;
      }

      if (controller is TargetController) {
        router.all(controller.path,
            controllerPipeline.addHandler((shelf.Request shelfRequest) async {
          // Create a Target request from the Shelf request.
          Request request = createRequest(
            this,
            shelfRequest,
          );

          // Handle the Target request to a Shelf request
          Future<shelf.Response> toShelfResponse(
            FutureOr<Response> response,
          ) async {
            try {
              Response handledResponse = await response;
              _assertNotResponseProxy(handledResponse);
              return handledResponse.delegate;
            } on RequestNotHandled {
              try {
                Response handledResponse = await await controller.any(request);
                _assertNotResponseProxy(handledResponse);
                return handledResponse.delegate;
              } on RequestNotHandled {
                return await _notFoundHandler(shelfRequest);
              } catch (error, stackTrace) {
                return await _serverErrorHandler(
                  shelfRequest,
                  error,
                  stackTrace,
                );
              }
            } catch (error, stackTrace) {
              return await _serverErrorHandler(shelfRequest, error, stackTrace);
            } finally {
              request.splitter.close();
            }
          }

          // A HEAD handler should always be implemeted if a GET handler is.
          // This function first tries to call the user defined HEAD handler,
          // or if it doesn't exist, then calls the GET handler but removes the
          // body from the response.
          Future<shelf.Response> handleHeadRequest(Request request) async {
            try {
              // First try to return a user implemented HEAD request.
              Response handledResponse = await controller.head(request);
              _assertNotResponseProxy(handledResponse);
              return handledResponse.delegate;
            } on RequestNotHandled {
              try {
                // If not implemented, try to call the GET handler
                final handledResponse = await controller.get(request);
                _assertNotResponseProxy(handledResponse);
                return _removeResponseBody(handledResponse.delegate);
              } on RequestNotHandled {
                // If no GET handler, throw a 404 as expected.
                return await _notFoundHandler(shelfRequest);
              } catch (error, stackTrace) {
                return await _serverErrorHandler(
                  shelfRequest,
                  error,
                  stackTrace,
                );
              }
            } catch (error, stackTrace) {
              return await _serverErrorHandler(
                shelfRequest,
                error,
                stackTrace,
              );
            } finally {
              request.splitter.close();
            }
          }

          try {
            final wsHandshake = shelfRequest.createWebSocketHandshake(
              configuration.ws,
            );

            // Only create a WebSocket handler if the handshake was a success.
            if (wsHandshake != null) {
              shelfRequest.createWebSocketHandler(wsHandshake, (handler) async {
                try {
                  final stream = controller.ws(request, handler);
                  await handler.delegate.sink.addStream(stream);
                  handler.close(
                    ws_status.normalClosure,
                    'The server closed the WebSocket handler.',
                  );
                } on RequestNotHandled {
                  handler.close(
                    ws_status.normalClosure,
                    'The server has no valid WebSocket handler.',
                  );
                } catch (_) {
                  // TODO(ehesp): Do something with the error (log it, event etc).
                  handler.close(ws_status.internalServerError);
                }
              });
            }
          } on WebSocketHandshakeAborted catch (e) {
            return shelf.Response.forbidden(e.reason);
          }

          switch (request.method) {
            case HttpMethod.connect:
              return toShelfResponse(controller.connect(request));
            case HttpMethod.delete:
              return toShelfResponse(controller.delete(request));
            case HttpMethod.get:
              return toShelfResponse(controller.get(request));
            case HttpMethod.head:
              return handleHeadRequest(request);
            case HttpMethod.options:
              return toShelfResponse(controller.options(request));
            case HttpMethod.patch:
              return toShelfResponse(controller.patch(request));
            case HttpMethod.post:
              return toShelfResponse(controller.post(request));
            case HttpMethod.put:
              return toShelfResponse(controller.put(request));

            default:
              return _notFoundHandler(shelfRequest);
          }
        }));
      }
    }

    final pipeline = _createServerPipeline();
    final staticHandler = _createStaticFileHandler();
    final routerHandler = pipeline.addHandler(router);

    shelf.Cascade cascade = shelf.Cascade();

    // First ensure static files are served before anything else
    if (staticHandler != null) {
      cascade = cascade.add(staticHandler);
    }

    // Add the controller route handler
    cascade = cascade.add(routerHandler);

    /// Create the HTTP server & start listening.
    server = await shelf_io.serve(
        cascade.handler, configuration.http.host, configuration.http.port);

    server!.autoCompress = configuration.http.compressed;

    return server!;
  }

  /// Stops the HttpServer.
  Future<void> stop({bool force = false}) async {
    if (server != null) {
      await server!.close(force: force);
      server = null;
    }
  }
}
