/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'dart:io';

import 'package:target_framework/server.dart';
import 'package:target_framework/target_framework.dart'
    show HttpMethod, TargetMiddleware, Response;
import 'package:target_framework/src/request.dart';
import 'package:target_framework/src/response.dart';
import 'package:shelf/shelf.dart' as shelf;

/// Returns a [HttpMethod] from a string.
HttpMethod stringToHttpMethod(String method) {
  switch (method.toLowerCase()) {
    case 'connect':
      return HttpMethod.connect;
    case 'delete':
      return HttpMethod.delete;
    case 'get':
      return HttpMethod.get;
    case 'head':
      return HttpMethod.head;
    case 'options':
      return HttpMethod.options;
    case 'patch':
      return HttpMethod.patch;
    case 'post':
      return HttpMethod.post;
    case 'put':
      return HttpMethod.put;
    default:
      throw new ArgumentError('Invalid HTTP method: $method');
  }
}

/// Creates a [shelf.Pipeline] from a list of [Middleware].
shelf.Pipeline createMiddlewarePipeline(shelf.Pipeline pipeline,
    TargetServer target, List<TargetMiddleware> middlewares) {
  for (final middleware in middlewares) {
    pipeline = pipeline.addMiddleware((shelf.Handler innerHandler) {
      // Handle the request
      return (shelf.Request shelfRequest) async {
        Request request = createRequest(
          target,
          shelfRequest,
        );

        // Pass the request to the middleware
        Response? response = await Future.sync(
          () async {
            // Call the middleware request handler
            Response? response = await middleware.requestHandler(request);

            // Ensure the internal shelf request is up-to-date
            shelfRequest = request.delegate;

            // Return the response
            return response;
          },
        );

        if (response != null) {
          return response.delegate;
        }

        // Call the inner handler (could be a next middleware or controller)
        shelf.Response shelfResponse = await Future.sync(() {
          return innerHandler(shelfRequest);
        });

        // Pass the response to the middleware response handler
        Response middlewareResponse = await middleware.responseHandler(
            request, Response.fromShelfResponse(shelfResponse));

        // Return the final response
        return middlewareResponse.delegate;
      };
    });
  }

  return pipeline;
}

const String? _targetApplicationDirectory =
    const bool.hasEnvironment('TARGET_APP_DIRECTORY')
        ? const String.fromEnvironment('TARGET_APP_DIRECTORY')
        : null;

/// Your application directory, e.g. `/Users/me/target-app/app`.
String get targetApplicationDirectory {
  return _targetApplicationDirectory ??
      Platform.environment['TARGET_APP_DIRECTORY'] ??
      Directory.current.path;
}

/// Whether or not the current Target server is running NOT running as a built `dart compile` executable.
const String targetEnvironment =
    const String.fromEnvironment('TARGET_ENV', defaultValue: 'development');

/// Whether or not the current Target server is running NOT running as a built `dart compile` executable.
const bool targetIsCompiled =
    const bool.fromEnvironment('TARGET_IS_COMPILED', defaultValue: false);
