/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'dart:io';
import 'dart:core';
import 'package:path/path.dart' as p;

import 'package:mustache_template/mustache_template.dart';
import 'utils.dart';

class View {
  View(this.path, this.data);

  /// The dotpath of the view file.
  final String path;

  /// Data to be passed to the view.
  ///
  /// This is rendered using Mustache syntax, e.g. {{ name }}
  final Map<String, dynamic> data;

  /// Converts the dotpath file name to a full file path.
  String dotPathToFilePath(String path) {
    return '${path.replaceAll('.', Platform.pathSeparator)}.mustache';
  }

  /// Gets the contents of a file from the view directory.
  String getFileContent(String path) {
    final file = File(
      p.join(targetApplicationDirectory, 'views', path),
    );

    if (!file.existsSync()) {
      throw Exception('File not found: ${file.path}');
    }

    return file.readAsStringSync();
  }

  /// Resolves a view file by dotpath.
  String _resolve(String dotPath, [String? yield]) {
    String html = '';

    // Get the full file path of the template
    final filePath = dotPathToFilePath(dotPath);

    // Get the raw template content
    final fileContent = getFileContent(filePath);

    String? extended;

    Template? Function(String name) resolver = (name) {
      if (name.startsWith('!extends ')) {
        // Template should only extend once.
        if (extended != null) {
          throw Exception('A layout may only extend once.');
        }

        // Get the dot path of the template to extend
        extended = name.replaceFirst('!extends ', '');

        // Return null since we want the extended template to include
        // the current template's content.
        return null;
      }

      // If there is a yielder and content to yield, yield it.
      if (name == '!yield' && yield != null) {
        return Template(yield, lenient: true);
      }

      // Otherwise attempt to resolve a user defined template.
      return Template(_resolve(name), lenient: true);
    };

    // Build the template instance with the current file content & resolver.
    final template = Template(fileContent,
        name: filePath, lenient: true, partialResolver: resolver);

    // Render the template to HTML.
    html = template.renderString(data);

    // If there is nothing to extend, just return the current HTML.
    if (extended == null) {
      return html;
    }

    // Otherwise, resolve the extended view with content to yield.
    return _resolve(extended!, html);
  }

  /// Builds the view with into an HTML string.
  String build() {
    return _resolve(path);
  }
}
