/*
 * Copyright (c) 2022-present Invertase Limited & Contributors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this library except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

import 'dart:async';
import 'dart:convert';

import 'package:shelf/shelf.dart' as shelf;
import 'package:target_framework/target_framework.dart';
import 'package:web_socket_channel/web_socket_channel.dart';

extension WebSocketExtension on shelf.Request {
  /// Creates a [WebSocketHandshake] if the incoming request is a websocket
  /// upgrade request.
  ///
  /// The request has to meet a certain number of conditions, such as being a
  /// GET request, including an "Origin" header, and having a "Sec-WebSocket-Key".
  ///
  /// The users [WebSocketConfiguration] is also checked to ensure that the request
  /// is from an allowed host.
  WebSocketHandshake? createWebSocketHandshake(WebSocketConfiguration config) {
    // Ignore any requests which don't look like a websocket request.
    if (method != 'GET') return null;
    if (headers['upgrade'] != 'websocket') return null;

    if (!canHijack) {
      throw WebSocketHandshakeAborted('Request cannot be hijacked.');
    }

    if (protocolVersion != '1.1') {
      throw WebSocketHandshakeAborted('Invalid protocol version header.');
    }

    if (headers['origin'] == null) {
      throw WebSocketHandshakeAborted('No origin header was found.');
    }

    // https://www.iana.org/assignments/websocket/websocket.xml#version-number
    if (headers['sec-websocket-version'] != '13') {
      throw WebSocketHandshakeAborted(
        'Invalid Sec-WebSocket-Version version header.',
      );
    }

    if (headers['sec-websocket-key'] == null) {
      throw WebSocketHandshakeAborted('No Sec-WebSocket-Key header was found.');
    }

    final origin = Uri.parse(headers['origin'] ?? '');
    final bool allowAnyOrigin = config.origins.contains(Uri(host: '*'));

    // TODO(ehesp) handle wildcard origins (e.g. 'http://*.example.com')
    if (!allowAnyOrigin && !config.origins.contains(origin)) {
      throw WebSocketHandshakeAborted(
        'Origin is not allowed for this WebSocket request.',
      );
    }

    String? protocol;
    final protocols = headers['sec-websocket-protocol'];
    if (protocols != null) {
      for (var el in protocols.split(',')) {
        el = el.trim();
        if (config.protocols.contains((o) => o == el)) {
          protocol = el;
          break;
        }
      }
    }

    return WebSocketHandshake._(
      key: WebSocketChannel.signKey(headers['sec-websocket-key']!),
      protocol: protocol,
      origin: origin.host,
      pingInterval: config.pingInterval,
    );
  }

  /// Creates a [WebSocketHandler] instance for the given [WebSocketHandshake].
  void createWebSocketHandler(
    WebSocketHandshake handshake,
    void Function(WebSocketHandler instance) cb,
  ) {
    hijack((channel) {
      final sink = utf8.encoder.startChunkedConversion(channel.sink)
        ..add(
          'HTTP/1.1 101 Switching Protocols\r\n'
          'Upgrade: websocket\r\n'
          'Connection: Upgrade\r\n'
          'Sec-WebSocket-Accept: ${handshake.key}\r\n',
        );

      if (handshake.protocol != null) {
        sink.add('Sec-WebSocket-Protocol: ${handshake.protocol}\r\n');
      }

      // https://developer.mozilla.org/en-US/docs/Web/API/WebSockets_API/Writing_WebSocket_servers#server_handshake_response
      sink.add('\r\n');

      final wsChannel = WebSocketChannel(
        channel,
        pingInterval: handshake.pingInterval,
        protocol: handshake.protocol,
      );

      cb(WebSocketHandler._(wsChannel));
    });
  }
}

/// The [WebSocketHandshake] class represents a successful handshake between
/// a client and server.
///
/// The handshake can be created by calling [createWebSocketHandshake].
class WebSocketHandshake {
  WebSocketHandshake._({
    required this.key,
    required this.origin,
    required this.protocol,
    required this.pingInterval,
  });

  /// The signed Sec-WebSocket-Key.
  final String key;

  /// The accepted origin of the request.
  final String? origin;
  final String? protocol;
  final Duration? pingInterval;
}

/// An extension to interact with the uderlying [WebSocketChannel].
///
/// This extension should not be publically exposed or accessed.
extension WebSocketHandlerExtension on WebSocketHandler {
  /// Returns the underlying [WebSocketChannel] instance.
  WebSocketChannel get delegate => _delegate;

  /// Closes the underlying [WebSocketSink].
  void close([int? code, String? reason]) => _delegate.sink.close(code, reason);
}

/// The open [WebSocketHandler] instance for a request.
class WebSocketHandler {
  WebSocketHandler._(this._delegate);

  final WebSocketChannel _delegate;

  Stream<Object?> get stream => _delegate.stream;
}
