import 'package:target_framework/target_framework.dart';
import 'package:spec/spec.dart';

import 'utils.dart';

void main() {
  group('$Cookie', () {
    group('parse', () {
      test('it returns null if no cookie is found', () async {
        final cookie = Cookie('foo');
        final request = FakeRequest(
          fakeHeaders: ImmutableHeaders({
            'Cookie': 'bar=baz',
          }),
        );

        final value = await cookie.parse(request);
        expect(value).isNull();
      });

      test('it returns the value if the cookie is found', () async {
        final cookie = Cookie('foo');
        final request = FakeRequest(
          fakeHeaders: ImmutableHeaders({
            'Cookie': 'foo=bar',
          }),
        );

        final value = await cookie.parse(request);
        expect(value).toEqual('bar');
      });

      test('it null if the cookie is not signed but has secrets', () async {
        final cookie = Cookie('foo', secrets: ['bar']);
        final request = FakeRequest(
          fakeHeaders: ImmutableHeaders({
            'Cookie': 'foo=bar',
          }),
        );

        final value = await cookie.parse(request);
        expect(value).isNull();
      });

      test('it decodes a signed cookie', () async {
        final cookie = Cookie('foo', secrets: ['bar']);
        final header = await cookie.serialize('baz');

        final request = FakeRequest(
          fakeHeaders: ImmutableHeaders({
            'Cookie': header,
          }),
        );

        final value = await cookie.parse(request);
        expect(value).toEqual('baz');
      });

      test('it retuns null if the signed cookie is not valid', () async {
        final cookie = Cookie('foo', secrets: ['bar']);
        final header = await cookie.serialize('baz');

        final request = FakeRequest(
          fakeHeaders: ImmutableHeaders({
            'Cookie': '$header-',
          }),
        );

        final value = await cookie.parse(request);
        expect(value).isNull();
      });
    });

    group('serialize', () {
      test('it serializes a value', () async {
        final cookie = Cookie('foo');
        await expect(cookie.serialize('bar')).completion.toEqual('foo=bar');
      });

      test('it serializes a value', () async {
        final cookie = Cookie('foo');
        await expect(cookie.serialize('bar')).completion.toEqual('foo=bar');
      });

      test('it serializes a value with a config domain', () async {
        final cookie = Cookie('foo', domain: 'baz');
        await expect(cookie.serialize('bar'))
            .completion
            .toEqual('foo=bar; Domain=baz');
      });

      test('it serializes a value with a override domain', () async {
        final cookie = Cookie('foo');
        await expect(cookie.serialize('bar', domain: 'baz'))
            .completion
            .toEqual('foo=bar; Domain=baz');
      });

      test('it serializes a value with a config expiry', () async {
        final cookie = Cookie('foo', expires: DateTime(2022, 2, 2));
        await expect(cookie.serialize('bar'))
            .completion
            .toEqual('foo=bar; Expires=Wed, 02 Feb 2022 00:00:00 GMT');
      });

      test('it serializes a value with a override expiry', () async {
        final cookie = Cookie('foo');
        await expect(cookie.serialize('bar', expires: DateTime(2022, 2, 2)))
            .completion
            .toEqual('foo=bar; Expires=Wed, 02 Feb 2022 00:00:00 GMT');
      });

      test('it serializes a value with a config http only', () async {
        final cookie = Cookie('foo', httpOnly: true);
        await expect(cookie.serialize('bar'))
            .completion
            .toEqual('foo=bar; HttpOnly');
      });

      test('it serializes a value with a override http only', () async {
        final cookie = Cookie('foo');
        await expect(cookie.serialize('bar', httpOnly: true))
            .completion
            .toEqual('foo=bar; HttpOnly');
      });

      test('it serializes a value with a config max age', () async {
        final cookie = Cookie('foo', maxAge: 10);
        await expect(cookie.serialize('bar'))
            .completion
            .toEqual('foo=bar; Max-Age=10');
      });

      test('it serializes a value with a override max age', () async {
        final cookie = Cookie('foo');
        await expect(cookie.serialize('bar', maxAge: 10))
            .completion
            .toEqual('foo=bar; Max-Age=10');
      });

      test('it serializes a value with a config path', () async {
        final cookie = Cookie('foo', path: '/foo');
        await expect(cookie.serialize('bar'))
            .completion
            .toEqual('foo=bar; Path=/foo');
      });

      test('it serializes a value with a override path', () async {
        final cookie = Cookie('foo');
        await expect(cookie.serialize('bar', path: '/foo'))
            .completion
            .toEqual('foo=bar; Path=/foo');
      });

      test('it serializes a value with a secure flag', () async {
        final cookie = Cookie('foo', secure: true);
        await expect(cookie.serialize('bar'))
            .completion
            .toEqual('foo=bar; Secure');
      });

      test('it serializes a value with a override secure flag', () async {
        final cookie = Cookie('foo');
        await expect(cookie.serialize('bar', secure: true))
            .completion
            .toEqual('foo=bar; Secure');
      });

      test('it serializes a value with a same site value', () async {
        final cookie1 = Cookie('foo', sameSite: CookieSameSite.lax);
        await expect(cookie1.serialize('bar'))
            .completion
            .toEqual('foo=bar; SameSite=Lax');

        final cookie2 = Cookie('foo', sameSite: CookieSameSite.none);
        await expect(cookie2.serialize('bar'))
            .completion
            .toEqual('foo=bar; SameSite=None');

        final cookie3 = Cookie('foo', sameSite: CookieSameSite.strict);
        await expect(cookie3.serialize('bar'))
            .completion
            .toEqual('foo=bar; SameSite=Strict');
      });

      test('it serializes a value with a override same site value', () async {
        final cookie1 = Cookie('foo');
        await expect(cookie1.serialize('bar', sameSite: CookieSameSite.lax))
            .completion
            .toEqual('foo=bar; SameSite=Lax');

        final cookie2 = Cookie('foo');
        await expect(cookie2.serialize('bar', sameSite: CookieSameSite.none))
            .completion
            .toEqual('foo=bar; SameSite=None');

        final cookie3 = Cookie('foo');
        await expect(cookie3.serialize('bar', sameSite: CookieSameSite.strict))
            .completion
            .toEqual('foo=bar; SameSite=Strict');
      });
    });

    test('isSigned returns the correct value', () {
      final cookie1 = Cookie('foo');
      expect(cookie1.isSigned).isFalse();

      final cookie2 = Cookie('foo', secrets: ['bar']);
      expect(cookie2.isSigned).isTrue();
    });
  });
}
