import 'package:mockito/mockito.dart';
import 'package:target_framework/target_framework.dart';
import 'package:spec/spec.dart';

import 'utils.dart';

void main() {
  group('$CorsMiddleware', () {
    test('it skips request handler if request is not OPTIONS', () async {
      final request = MockRequest();
      final middleware = CorsMiddleware();

      final res = await middleware.requestHandler(request);

      expect(res).isNull();
    });

    group('origin', () {
      test('it does not apply an origin if not specified', () async {
        final request = MockRequest();
        final response = MockResponse();
        final middleware = CorsMiddleware();

        final res = await middleware.responseHandler(request, response);

        expect(res.headers[HttpHeaders.accessControlAllowOriginHeader])
            .isNull();
      });

      test('it applies an origin if specified', () async {
        final request = MockRequest();
        final response = MockResponse();
        final middleware = CorsMiddleware(origin: 'foo');

        final res = await middleware.responseHandler(request, response);

        expect(res.headers[HttpHeaders.accessControlAllowOriginHeader])
            .toEqual('foo');
      });

      test('it applies a * origin if specified', () async {
        final request = MockRequest();
        final response = MockResponse();
        final middleware = CorsMiddleware(origin: '*');

        final res = await middleware.responseHandler(request, response);

        expect(res.headers[HttpHeaders.accessControlAllowOriginHeader])
            .toEqual('*');
      });

      test('it allows configuring an origin', () async {
        final request = MockRequest();
        final response = MockResponse();
        final middleware =
            CorsMiddleware(origin: '*', configureOrigin: (request) => 'foo');

        final res = await middleware.responseHandler(request, response);

        expect(res.headers[HttpHeaders.accessControlAllowOriginHeader])
            .toEqual('foo');
      });

      test('it skips configuring an origin if null is returned', () async {
        final request = MockRequest();
        final response = MockResponse();
        final middleware =
            CorsMiddleware(origin: '*', configureOrigin: (request) => null);

        final res = await middleware.responseHandler(request, response);

        expect(res.headers[HttpHeaders.accessControlAllowOriginHeader])
            .isNull();
      });
    });

    group('credentials', () {
      test('it skips applying credentials if not true', () async {
        final request = MockRequest();
        final response = MockResponse();
        final middleware = CorsMiddleware();

        final res = await middleware.responseHandler(request, response);

        expect(res.headers[HttpHeaders.accessControlAllowCredentialsHeader])
            .isNull();
      });

      test('it applies credentials if true', () async {
        final request = MockRequest();
        final response = MockResponse();
        final middleware = CorsMiddleware(credentials: true);

        final res = await middleware.responseHandler(request, response);

        expect(res.headers[HttpHeaders.accessControlAllowCredentialsHeader])
            .toEqual('true');
      });
    });

    group('exposed headers', () {
      test('it skips applying exposed headers if not provided', () async {
        final request = MockRequest();
        final response = MockResponse();
        final middleware = CorsMiddleware();

        final res = await middleware.responseHandler(request, response);

        expect(res.headers[HttpHeaders.accessControlExposeHeadersHeader])
            .isNull();
      });

      test('it applies exposed headers if provided', () async {
        final request = MockRequest();
        final response = MockResponse();
        final middleware = CorsMiddleware(exposedHeaders: ['foo', 'bar']);

        final res = await middleware.responseHandler(request, response);

        expect(res.headers[HttpHeaders.accessControlExposeHeadersHeader])
            .toEqual('foo, bar');
      });
    });

    group('methods', () {
      test('it applies default methods', () async {
        final request = MockRequest();
        final middleware = CorsMiddleware();

        when(request.method).thenReturn(HttpMethod.options);

        final res = await middleware.requestHandler(request);

        expect(res!.headers[HttpHeaders.accessControlAllowMethodsHeader])
            .toEqual('GET, HEAD, PUT, PATCH, POST, DELETE');
      });

      test('it applies custom methods', () async {
        final request = MockRequest();
        final middleware =
            CorsMiddleware(methods: [HttpMethod.get, HttpMethod.delete]);

        when(request.method).thenReturn(HttpMethod.options);

        final res = await middleware.requestHandler(request);

        expect(res!.headers[HttpHeaders.accessControlAllowMethodsHeader])
            .toEqual('GET, DELETE');
      });

      test('it doesnt apply methods if empty', () async {
        final request = MockRequest();
        final middleware = CorsMiddleware(methods: []);

        when(request.method).thenReturn(HttpMethod.options);

        final res = await middleware.requestHandler(request);

        expect(res!.headers[HttpHeaders.accessControlAllowMethodsHeader])
            .isNull();
      });
    });

    group('allowed headers', () {
      test('it applies provided headers', () async {
        final request = MockRequest();
        final middleware = CorsMiddleware(allowedHeaders: ['foo', 'bar']);

        when(request.method).thenReturn(HttpMethod.options);

        final res = await middleware.requestHandler(request);

        expect(res!.headers[HttpHeaders.accessControlAllowHeadersHeader])
            .toEqual('foo, bar');
      });

      test('it applies requested headers if no allowed headers are set',
          () async {
        final request = MockRequest();
        final middleware = CorsMiddleware();

        when(request.method).thenReturn(HttpMethod.options);
        when(request.headers).thenReturn(ImmutableHeaders({
          HttpHeaders.accessControlRequestHeadersHeader: 'foo, bar',
        }));

        final res = await middleware.requestHandler(request);

        expect(res!.headers[HttpHeaders.accessControlAllowHeadersHeader])
            .toEqual('foo, bar');
        expect(res.headers[HttpHeaders.varyHeader])
            .toEqual(HttpHeaders.accessControlRequestHeadersHeader);
      });

      test(
          'it doesnt apply requested headers if no allowed headers are set or requested',
          () async {
        final request = MockRequest();
        final middleware = CorsMiddleware();

        when(request.method).thenReturn(HttpMethod.options);

        final res = await middleware.requestHandler(request);

        expect(res!.headers[HttpHeaders.accessControlAllowHeadersHeader])
            .isNull();
      });
    });

    group('max age', () {
      test('it applies a max age if provided', () async {
        final request = MockRequest();
        final middleware = CorsMiddleware(maxAge: 1234);

        when(request.method).thenReturn(HttpMethod.options);

        final res = await middleware.requestHandler(request);

        expect(res!.headers[HttpHeaders.accessControlMaxAgeHeader])
            .toEqual('1234');
      });

      test('it doesnt apply a max age if not provided', () async {
        final request = MockRequest();
        final middleware = CorsMiddleware();

        when(request.method).thenReturn(HttpMethod.options);

        final res = await middleware.requestHandler(request);

        expect(res!.headers[HttpHeaders.accessControlMaxAgeHeader]).isNull();
      });
    });

    group('options success status', () {
      test('if applies a custom status code', () async {
        final request = MockRequest();
        final middleware = CorsMiddleware(optionsSuccessStatus: 200);

        when(request.method).thenReturn(HttpMethod.options);

        final res = await middleware.requestHandler(request);

        expect(res == null).isFalse();
        expect(res!.status).toEqual(200);
      });
    });
  });
}
