import 'package:target_framework/target_framework.dart';
import 'package:spec/spec.dart';

void main() {
  group('$Headers', () {
    test('it accepts initial headers', () {
      final initial = {'foo': 'bar'};
      final h = Headers(initial);
      expect(h.toMap()).toEqual(initial);
    });

    test('it clones headers instance (.clone)', () {
      final initial = {'foo': 'bar'};
      final h1 = Headers(initial);
      final h2 = h1.clone();
      h2.append('foo', 'baz');

      expect(h2.toMap()).toEqual({'foo': 'bar, baz'});
      expect(h1.toMap()).toEqual(initial);
    });

    test('it clones headers instance (.fromHeaders)', () {
      final initial = {'foo': 'bar'};
      final h1 = Headers(initial);
      final h2 = Headers.fromHeaders(h1);
      h2.append('foo', 'baz');

      expect(h2.toMap()).toEqual({'foo': 'bar, baz'});
      expect(h1.toMap()).toEqual(initial);
    });

    test('append adds a new header if none exists', () {
      final h1 = Headers();
      h1.append('foo', 'bar');

      expect(h1.toMap()).toEqual({'foo': 'bar'});
    });

    test('append adds a existing header if exists', () {
      final h1 = Headers();
      h1.append('foo', 'bar');
      h1.append('foo', 'baz');

      expect(h1.toMap()).toEqual({'foo': 'bar, baz'});
    });

    test('append works case insensitive', () {
      final h1 = Headers();
      h1.append('foo', 'bar');
      h1.append('FOO', 'BAR');

      // Field values are not case insensitive
      expect(h1.toMap()).toEqual({'foo': 'bar, BAR'});
    });

    test('delete removes a header', () {
      final h1 = Headers();
      h1.append('foo', 'bar');
      h1.delete('foo');

      expect(h1.toMap()).toEqual({});
    });

    test('delete removes a header (case insensitive)', () {
      final h1 = Headers();
      h1.append('foo', 'bar');
      h1.delete('FOO');

      expect(h1.toMap()).toEqual({});
    });

    test('has returns the correct boolean', () {
      final h1 = Headers();
      h1.append('foo', 'bar');

      expect(h1.has('foo')).isTrue();
      expect(h1.has('FOO')).isTrue();
      expect(h1.has('bar')).isFalse();
      expect(h1.has('BAR')).isFalse();
    });

    test('keys returns the correct header keys', () {
      final h1 = Headers();
      h1.append('foo', 'bar');
      h1.append('bar', 'baz');
      h1.append('BAZ', 'ben');

      expect(h1.keys.toList()).toEqual(['foo', 'bar', 'baz']);
    });

    test('keys returns the correct header values', () {
      final h1 = Headers();
      h1.append('foo', 'bar');
      h1.append('FOO', 'baz');
      h1.append('bar', 'baz');
      h1.append('BAZ', 'ben');

      expect(h1.values.toList()).toEqual(['bar, baz', 'baz', 'ben']);
    });

    test('overrides the get operator', () {
      final h1 = Headers();
      h1.append('foo', 'bar');
      h1.append('bar', 'baz');

      expect(h1['foo']).toEqual('bar');
      expect(h1['FoO']).toEqual('bar');
      expect(h1['bar']).toEqual('baz');
      expect(h1['baR']).toEqual('baz');

      expect(h1['nope']).isNull();
    });

    test('overrides the set operator', () {
      final h1 = Headers();
      h1.append('foo', 'bar');
      h1.append('foo', 'baz');
      h1.append('bar', 'baz');

      h1['foo'] = 'ben';
      h1['BAR'] = 'foo';
      h1['baz'] = 'bar';

      expect(h1['foo']).toEqual('ben');
      expect(h1['bar']).toEqual('foo');
      expect(h1['baz']).toEqual('bar');
    });
  });
}
