import 'package:mockito/mockito.dart';
import 'package:spec/spec.dart';
import 'package:target_framework/target_framework.dart';
import 'utils.dart';

final int port = 3005;

void main() {
  group('Server Controller Middleware', () {
    group('requestHandler', () {
      test('should call middleware handler', () async {
        final mock = RequestHandlerMock();

        await startServer(port, controllers: [
          TestController('/foo',
              middleware: [TestMiddleware(handleRequest: mock)],
              anyHandler: (req) async {
            return Response('bar');
          })
        ]);

        final req = await request(HttpMethod.get, '/foo');

        verify(mock(any)).called(1);
        verifyNoMoreInteractions(mock);

        expect(req.statusCode).toEqual(200);
        expect(req.body).toEqual('bar');
      });

      test('should fall through if null is returned', () async {
        final mock1 = RequestHandlerMock();
        final mock2 = RequestHandlerMock();

        await startServer(port, controllers: [
          TestController('/foo', middleware: [
            TestMiddleware(handleRequest: mock1),
            TestMiddleware(handleRequest: mock2)
          ], anyHandler: (req) async {
            return Response('bar');
          })
        ]);

        final req = await request(HttpMethod.get, '/foo');

        verifyInOrder([mock1(any), mock2(any)]);
        verifyNoMoreInteractions(mock1);
        verifyNoMoreInteractions(mock2);

        expect(req.statusCode).toEqual(200);
        expect(req.body).toEqual('bar');
      });

      test('should not fallthrough if a request is returned', () async {
        final mock1 = RequestHandlerMock();
        final mock2 = RequestHandlerMock();

        when(mock1(any))
            .thenAnswer((_) => Future.value(Response.badRequest('foo')));

        await startServer(port, controllers: [
          TestController('/foo', middleware: [
            TestMiddleware(handleRequest: mock1),
            TestMiddleware(handleRequest: mock2)
          ], anyHandler: (req) async {
            return Response('bar');
          })
        ]);

        final res = await request(HttpMethod.get, '/foo');

        verify(mock1(any)).called(1);
        verifyNoMoreInteractions(mock1);
        verifyZeroInteractions(mock2);

        expect(res.statusCode).toEqual(400);
        expect(res.body).toEqual('foo');
      });
    });

    group('responseHandler', () {
      test('should call middleware handler', () async {
        final mock = ResponseHandlerMock();

        when(mock(any, any))
            .thenAnswer((_) => Future.value(Response.ok('foo')));

        await startServer(port, controllers: [
          TestController('/foo', middleware: [
            TestMiddleware(handleResponse: mock),
          ], anyHandler: (req) async {
            return Response('bar');
          })
        ]);

        final req = await request(HttpMethod.get, '/foo');

        verify(mock(any, any)).called(1);
        verifyNoMoreInteractions(mock);

        expect(req.statusCode).toEqual(200);
        expect(req.body).toEqual('foo');
      });

      test('should fall through in reverse order', () async {
        final mock1 = ResponseHandlerMock();
        final mock2 = ResponseHandlerMock();

        when(mock1(any, any))
            .thenAnswer((_) => Future.value(Response.ok('foo')));
        when(mock2(any, any))
            .thenAnswer((_) => Future.value(Response.ok('bar')));

        await startServer(port, controllers: [
          TestController('/foo', middleware: [
            TestMiddleware(handleResponse: mock1),
            TestMiddleware(handleResponse: mock2),
          ], anyHandler: (req) async {
            return Response('bar');
          })
        ]);

        final res = await request(HttpMethod.get, '/foo');

        // The response handlers are called in reverse order so the first
        // middleware is called last.
        verifyInOrder([mock2(any, any), mock1(any, any)]);

        verifyNoMoreInteractions(mock1);
        verifyNoMoreInteractions(mock2);

        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('foo');
      });
    });
  });
}
