import 'package:spec/spec.dart';
import 'package:target_framework/target_framework.dart';
import 'utils.dart';

final int port = 3000;

void main() {
  group('HttpMethod', () {
    group('ANY', () {
      test('should match any method', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.ok(req.method.value.toUpperCase());
          })
        ]);
        await request(HttpMethod.get, '/foo')
            .then((res) => expect(res.body).toEqual('GET'));
        await request(HttpMethod.post, '/foo')
            .then((res) => expect(res.body).toEqual('POST'));
        await request(HttpMethod.delete, '/foo')
            .then((res) => expect(res.body).toEqual('DELETE'));
        await request(HttpMethod.patch, '/foo')
            .then((res) => expect(res.body).toEqual('PATCH'));
        await request(HttpMethod.put, '/foo')
            .then((res) => expect(res.body).toEqual('PUT'));
      });

      test('should ignore a method if it is defined', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.ok('ANY');
          }, postHandler: (req) async {
            return Response.ok('POST');
          })
        ]);
        await request(HttpMethod.post, '/foo')
            .then((res) => expect(res.body).toEqual('POST'));
        await request(HttpMethod.get, '/foo')
            .then((res) => expect(res.body).toEqual('ANY'));
      });
    });

    group('DELETE', () {
      test('Returns an expected 200 text response', () async {
        await startServer(port, controllers: [
          TestController('/foo', deleteHandler: (req) async {
            return Response.ok(req.method.value.toUpperCase());
          })
        ]);
        final res = await request(HttpMethod.delete, '/foo');
        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('DELETE');
      });

      test('Returns an expected 404 when no handler', () async {
        await startServer(port, controllers: [TestController('/foo')]);
        final res = await request(HttpMethod.delete, '/foo');
        expect(res.statusCode).toEqual(404);
      });
    });

    group('GET', () {
      test('Returns an expected 200 text response', () async {
        await startServer(port, controllers: [
          TestController('/foo', getHandler: (req) async {
            return Response.ok(req.method.value.toUpperCase());
          })
        ]);
        final res = await request(HttpMethod.get, '/foo');
        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('GET');
      });

      test('Returns an expected 404 when no handler', () async {
        await startServer(port, controllers: [TestController('/foo')]);
        final res = await request(HttpMethod.get, '/foo');
        expect(res.statusCode).toEqual(404);
      });
    });

    group('HEAD', () {
      test('should 404 if not implemented', () async {
        await startServer(port, controllers: [TestController('/foo')]);
        await request(HttpMethod.head, '/foo')
            .then((res) => expect(res.statusCode).toEqual(404));
      });

      test('should return a response if implemented', () async {
        await startServer(port, controllers: [
          TestController('/foo', headHandler: (req) async {
            return Response.ok('123456789');
          })
        ]);
        await request(HttpMethod.head, '/foo').then((res) {
          expect(res.statusCode).toEqual(200);
          expect(res.body)
              .toEqual(''); // HEAD requests should not return a body
          expect(res.headers['content-length']).toEqual('9');
        });
      });

      test('should delegate to GET handler if implemented', () async {
        await startServer(port, controllers: [
          TestController('/foo', getHandler: (req) async {
            return Response.ok('123456789');
          })
        ]);
        await request(HttpMethod.head, '/foo').then((res) {
          expect(res.statusCode).toEqual(200);
          expect(res.body)
              .toEqual(''); // HEAD requests should not return a body
          expect(res.headers['content-length']).toEqual('9');
        });
      });
    });

    group('PATCH', () {
      test('Returns an expected 200 text response', () async {
        await startServer(port, controllers: [
          TestController('/foo', patchHandler: (req) async {
            return Response.ok(req.method.value.toUpperCase());
          })
        ]);
        final res = await request(HttpMethod.patch, '/foo');
        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('PATCH');
      });

      test('Returns an expected 404 when no handler', () async {
        await startServer(port, controllers: [TestController('/foo')]);
        final res = await request(HttpMethod.patch, '/foo');
        expect(res.statusCode).toEqual(404);
      });
    });

    group('POST', () {
      test('Returns an expected 200 text response', () async {
        await startServer(port, controllers: [
          TestController('/foo', postHandler: (req) async {
            return Response.ok(req.method.value.toUpperCase());
          })
        ]);
        final res = await request(HttpMethod.post, '/foo');
        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('POST');
      });

      test('Returns an expected 404 when no handler', () async {
        await startServer(port, controllers: [TestController('/foo')]);
        final res = await request(HttpMethod.post, '/foo');
        expect(res.statusCode).toEqual(404);
      });
    });

    group('PUT', () {
      test('Returns an expected 200 text response', () async {
        await startServer(port, controllers: [
          TestController('/foo', putHandler: (req) async {
            return Response.ok(req.method.value.toUpperCase());
          })
        ]);
        final res = await request(HttpMethod.put, '/foo');
        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('PUT');
      });

      test('Returns an expected 404 when no handler', () async {
        await startServer(port, controllers: [TestController('/foo')]);
        final res = await request(HttpMethod.put, '/foo');
        expect(res.statusCode).toEqual(404);
      });
    });
  });
}
