import 'dart:convert';

import 'package:target_framework/target_framework.dart';
import 'package:spec/spec.dart';

import 'utils.dart';

final int port = 3009;

void main() {
  group('$Request', () {
    test('can apply & read context', () async {
      final key = RequestContextKey();
      await startServer(port, controllers: [
        TestController('/foo', middleware: [
          TestMiddleware(handleRequest: (req) async {
            req.attach(key, 'bar');
            return null;
          }),
        ], anyHandler: (req) async {
          return Response(req.get(key));
        })
      ]);

      await request(HttpMethod.get, '/foo').then((res) {
        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('bar');
      });
    });

    test('can returns default context', () async {
      final key = RequestContextKey((r) => 'bar');
      await startServer(port, controllers: [
        TestController('/foo', anyHandler: (req) async {
          return Response(req.get(key));
        })
      ]);

      await request(HttpMethod.get, '/foo').then((res) {
        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('bar');
      });
    });

    test('it throws if no value is provided', () async {
      final key = RequestContextKey<String>();
      await startServer(port, controllers: [
        TestController('/foo', anyHandler: (req) async {
          final value = req.get(key);
          return Response(value);
        })
      ]);

      await request(HttpMethod.get, '/foo').then((res) {
        expect(res.statusCode).toEqual(500);
      });
    });

    test('it allows header overrides', () async {
      await startServer(port, controllers: [
        TestController('/foo', anyHandler: (req) async {
          req.change(headers: Headers({'foo': 'bar'}));
          print(req.headers);
          return Response.json(req.headers.toMap());
        })
      ]);

      await request(HttpMethod.get, '/foo').then((res) {
        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('{"foo":"bar"}');
      });
    });

    test('it allows header appending', () async {
      await startServer(port, controllers: [
        TestController('/foo', anyHandler: (req) async {
          req.change(headers: req.headers.clone()..append('foo', 'bar'));
          return Response.json(req.headers.toMap());
        })
      ]);

      await request(HttpMethod.get, '/foo').then((res) {
        expect(res.statusCode).toEqual(200);
        Map json = jsonDecode(res.body) as Map;
        expect(json['foo']).toEqual('bar');
        expect(json.keys.length).greaterThan(1);
      });
    });

    test('it allows body changing', () async {
      await startServer(port, controllers: [
        TestController('/foo', anyHandler: (req) async {
          req.change(body: utf8.encode('bar'));
          final body = await req.text();
          return Response.ok(body);
        })
      ]);

      await request(HttpMethod.post, '/foo', body: 'foo').then((res) {
        print(res.body);
        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('bar');
      });
    });
  });
}
