import 'dart:convert';

import 'package:spec/spec.dart';
import 'package:target_framework/target_framework.dart';
import 'utils.dart';

final int port = 3001;

void main() {
  group('Response', () {
    test('should return 200 with a string value', () async {
      await startServer(port, controllers: [
        TestController('/foo', anyHandler: (req) async {
          return Response('bar');
        })
      ]);
      await request(HttpMethod.get, '/foo').then((res) {
        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('bar');
      });
    });

    test('should return a custom status code', () async {
      await startServer(port, controllers: [
        TestController('/foo', anyHandler: (req) async {
          return Response('bar', status: 404);
        })
      ]);
      await request(HttpMethod.get, '/foo').then((res) {
        expect(res.statusCode).toEqual(404);
        expect(res.body).toEqual('bar');
      });
    });

    test('should return a custom headers', () async {
      await startServer(port, controllers: [
        TestController('/foo', anyHandler: (req) async {
          return Response('bar', headers: Headers()..append('x-foo', 'bar'));
        })
      ]);
      await request(HttpMethod.get, '/foo').then((res) {
        expect(res.statusCode).toEqual(200);
        expect(res.headers['x-foo']).toEqual('bar');
      });
    });

    group('ok', () {
      test('should return 200 with a string value', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.ok('bar');
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(200);
          expect(res.body).toEqual('bar');
        });
      });

      test('should return 200 with no body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.ok(null);
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(200);
          expect(res.body).toEqual('');
        });
      });

      test('should allow headers to be set', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.ok(null,
                headers: Headers()..append('x-foo', 'bar'));
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(200);
          expect(res.headers['x-foo']).toEqual('bar');
        });
      });
    });

    group('json', () {
      test('should return 200 with a map value', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.json({'foo': 'bar'});
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(200);
          final json = jsonDecode(res.body);
          expect(json['foo']).toEqual('bar');
        });
      });

      test('should return 200 with a correct content-type', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.json({'foo': 'bar'});
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(200);
          expect(res.headers['content-type'])
              .toEqual('application/json; charset=utf-8');
        });
      });

      test('should override the content-type if set', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.json(
              {'foo': 'bar'},
              headers: Headers()..append('content-type', 'text/html'),
            );
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(200);
          expect(res.headers['content-type'])
              .toEqual('application/json; charset=utf-8');
        });
      });

      test('should allow setting of a status code', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.json(
              {'foo': 'bar'},
              status: 404,
            );
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(404);
        });
      });
    });

    group('badRequest', () {
      test('should return 400 with no body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.badRequest(null);
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(400);
          expect(res.body).toEqual('');
        });
      });

      test('should return 400 with a body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.badRequest('foo');
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(400);
          expect(res.body).toEqual('foo');
        });
      });

      test('should allow headers to be set', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.badRequest(null,
                headers: Headers()..append('x-foo', 'bar'));
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(400);
          expect(res.headers['x-foo']).toEqual('bar');
        });
      });
    });

    group('unauthorized', () {
      test('should return 401 with no body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.unauthorized(null);
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(401);
          expect(res.body).toEqual('');
        });
      });

      test('should return 401 with a body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.unauthorized('foo');
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(401);
          expect(res.body).toEqual('foo');
        });
      });

      test('should allow headers to be set', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.unauthorized(null,
                headers: Headers()..append('x-foo', 'bar'));
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(401);
          expect(res.headers['x-foo']).toEqual('bar');
        });
      });
    });

    group('forbidden', () {
      test('should return 403 with no body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.forbidden(null);
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(403);
          expect(res.body).toEqual('');
        });
      });

      test('should return 403 with a body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.forbidden('foo');
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(403);
          expect(res.body).toEqual('foo');
        });
      });

      test('should allow headers to be set', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.forbidden(null,
                headers: Headers()..append('x-foo', 'bar'));
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(403);
          expect(res.headers['x-foo']).toEqual('bar');
        });
      });
    });

    group('notFound', () {
      test('should return 404 with no body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.notFound(null);
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(404);
          expect(res.body).toEqual('');
        });
      });

      test('should return 404 with a body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.notFound('foo');
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(404);
          expect(res.body).toEqual('foo');
        });
      });

      test('should allow headers to be set', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.notFound(null,
                headers: Headers()..append('x-foo', 'bar'));
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(404);
          expect(res.headers['x-foo']).toEqual('bar');
        });
      });
    });

    group('serverError', () {
      test('should return 500 with no body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.serverError(null);
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(500);
          expect(res.body).toEqual('');
        });
      });

      test('should return 500 with a body', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.serverError('foo');
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(500);
          expect(res.body).toEqual('foo');
        });
      });

      test('should allow headers to be set', () async {
        await startServer(port, controllers: [
          TestController('/foo', anyHandler: (req) async {
            return Response.serverError(null,
                headers: Headers()..append('x-foo', 'bar'));
          })
        ]);
        await request(HttpMethod.get, '/foo').then((res) {
          expect(res.statusCode).toEqual(500);
          expect(res.headers['x-foo']).toEqual('bar');
        });
      });
    });
  });
}
