import 'package:mockito/mockito.dart';
import 'package:spec/spec.dart';
import 'package:target_framework/target_framework.dart';
import 'utils.dart';

final int port = 3004;

void main() {
  group('Server Middleware', () {
    group('requestHandler', () {
      test('should call middleware handler', () async {
        final mock = RequestHandlerMock();

        final config = TestConfig(
          httpConfig: HttpConfiguration(
            middleware: [TestMiddleware(handleRequest: mock)],
          ),
        );

        await startServer(port, config: config);
        await request(HttpMethod.get, '/');

        verify(mock(any)).called(1);
        verifyNoMoreInteractions(mock);
      });

      test('should fall through if null is returned', () async {
        final mock1 = RequestHandlerMock();
        final mock2 = RequestHandlerMock();

        final config = TestConfig(
          httpConfig: HttpConfiguration(
            middleware: [
              TestMiddleware(handleRequest: mock1),
              TestMiddleware(handleRequest: mock2)
            ],
          ),
        );

        await startServer(port, config: config);
        final res = await request(HttpMethod.get, '/');

        verifyInOrder([mock1(any), mock2(any)]);
        verifyNoMoreInteractions(mock1);
        verifyNoMoreInteractions(mock2);

        expect(res.statusCode).toEqual(404);
      });

      test('should not fallthrough if a request is returned', () async {
        final mock1 = RequestHandlerMock();
        final mock2 = RequestHandlerMock();

        when(mock1(any)).thenAnswer((_) => Future.value(Response.ok('foo')));

        final config = TestConfig(
          httpConfig: HttpConfiguration(
            middleware: [
              TestMiddleware(handleRequest: mock1),
              TestMiddleware(handleRequest: mock2)
            ],
          ),
        );

        await startServer(port, config: config);
        final res = await request(HttpMethod.get, '/');

        verify(mock1(any)).called(1);
        verifyNoMoreInteractions(mock1);
        verifyZeroInteractions(mock2);

        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('foo');
      });
    });

    group('responseHandler', () {
      test('should call middleware handler', () async {
        final mock = ResponseHandlerMock();

        when(mock(any, any))
            .thenAnswer((_) => Future.value(Response.ok('foo')));

        final config = TestConfig(
          httpConfig: HttpConfiguration(
            middleware: [TestMiddleware(handleResponse: mock)],
          ),
        );

        await startServer(port, config: config);
        final req = await request(HttpMethod.get, '/');

        verify(mock(any, any)).called(1);
        verifyNoMoreInteractions(mock);

        expect(req.statusCode).toEqual(200);
        expect(req.body).toEqual('foo');
      });

      test('should fall through in reverse order', () async {
        final mock1 = ResponseHandlerMock();
        final mock2 = ResponseHandlerMock();

        when(mock1(any, any))
            .thenAnswer((_) => Future.value(Response.ok('foo')));
        when(mock2(any, any))
            .thenAnswer((_) => Future.value(Response.ok('bar')));

        final config = TestConfig(
          httpConfig: HttpConfiguration(
            middleware: [
              TestMiddleware(handleResponse: mock1),
              TestMiddleware(handleResponse: mock2),
            ],
          ),
        );

        await startServer(port, config: config);
        final res = await request(HttpMethod.get, '/');

        // The response handlers are called in reverse order so the first
        // middleware is called last.
        verifyInOrder([mock2(any, any), mock1(any, any)]);

        verifyNoMoreInteractions(mock1);
        verifyNoMoreInteractions(mock2);

        expect(res.statusCode).toEqual(200);
        expect(res.body).toEqual('foo');
      });
    });
  });
}
