import 'dart:io';

import 'package:async/async.dart';
import 'package:target_framework/target_framework.dart';
import 'package:spec/spec.dart';
import 'package:test/test.dart' show fail;

import 'utils.dart';

final int port = 3006;

void main() {
  group('Server WebSocket', () {
    WebSocket? ws;

    tearDown(() async {
      if (ws != null) {
        await ws!.close();
        ws = null;
      }
    });

    test('should not connect as no origin was allowed (default behavior)',
        () async {
      await startServer(port, controllers: []);

      // TODO handle with spec when new matcher
      try {
        await WebSocket.connect('ws://localhost:$port/foo');
        fail('Should not connect');
      } on WebSocketException catch (e) {
        expect(e.toString().contains('was not upgraded')).isTrue();
      } catch (e) {
        fail(e.toString());
      }
    });

    test('should not connect to the websocket server if the origin is invalid',
        () async {
      await startServer(
        port,
        config: TestConfig(
          wsConfig: WebSocketConfiguration(origins: ['foobar']),
        ),
        controllers: [
          TestController('/foo', wsHandler: (request, ws) async* {})
        ],
      );

      try {
        await WebSocket.connect(
          'ws://localhost:$port/foo',
          headers: {'Origin': 'foo'},
        );
        fail('Should not connect');
      } on WebSocketException catch (e) {
        expect(e.toString().contains('was not upgraded')).isTrue();
      } catch (e) {
        fail(e.toString());
      }
    });

    test('should connect to the websocket server', () async {
      await startServer(
        port,
        config: TestConfig(wsConfig: WebSocketConfiguration(origins: ['*'])),
        controllers: [
          TestController('/foo', wsHandler: (request, ws) async* {})
        ],
      );

      ws = await WebSocket.connect(
        'ws://localhost:$port/foo',
        headers: {'origin': 'localhost'},
      );
    });

    test('should recieve messages from the server', () async {
      await startServer(
        port,
        config: TestConfig(wsConfig: WebSocketConfiguration(origins: ['*'])),
        controllers: [
          TestController('/foo', wsHandler: (request, ws) async* {
            await for (var message in ws.stream) {
              if (message is String) {
                yield '$message - bar';
              }
            }
          })
        ],
      );

      ws = await WebSocket.connect(
        'ws://localhost:$port/foo',
        headers: {'origin': 'localhost'},
      );

      final stream = StreamQueue(ws!);

      ws!.add('foo');
      await expect(stream.next).completion.toEqual('foo - bar');

      ws!.add('123');
      await expect(stream.next).completion.toEqual('123 - bar');
    });
  });
}
