import 'dart:async';
import 'dart:io' show HttpServer;

import 'package:mockito/mockito.dart';
import 'package:target_framework/src/view.dart';
import 'package:target_framework/target_framework.dart';
import 'package:target_framework/server.dart';
import 'package:shelf/shelf.dart' as shelf;
import 'package:http/http.dart' as http;
import 'package:test/scaffolding.dart';

HttpServer? server;
http.Client _client = http.Client();

class DefaultConfig extends TargetConfiguration {
  DefaultConfig(this.port, [this.customConfig]);
  final int port;
  final TargetConfiguration? customConfig;

  @override
  HttpConfiguration get http {
    return HttpConfiguration(
      port: port,
      compressed: customConfig?.http.compressed ?? true,
      middleware: customConfig?.http.middleware ?? [],
    );
  }

  @override
  WebSocketConfiguration get ws => customConfig?.ws ?? WebSocketConfiguration();
}

/// Starts a server with config and controllers.
Future<void> startServer(
  int port, {
  TargetConfiguration? config,
  List<TargetBaseController>? controllers,
}) async {
  TargetServer instance = TargetServer(
    configuration: DefaultConfig(port, config),
    controllers: controllers ?? [],
  );

  // Ensure the server is stopped after each test.
  addTearDown(() async {
    await instance.stop();
  });

  server = await instance.start();
}

/// Performs a HTTP request to the server and returns the response.
Future<http.Response> request(
  HttpMethod method,
  String path, {
  Headers? headers,
  dynamic body,
}) async {
  final url = Uri.http('${server!.address.host}:${server!.port}', path);

  switch (method) {
    case HttpMethod.connect:
      throw Exception('CONNECT not supported');
    case HttpMethod.delete:
      return _client.delete(url, headers: headers?.toMap(), body: body);
    case HttpMethod.get:
      return _client.get(url, headers: headers?.toMap());
    case HttpMethod.head:
      return _client.head(url, headers: headers?.toMap());
    case HttpMethod.options:
      throw Exception('OPTIONS not supported');
    case HttpMethod.patch:
      return _client.patch(url, headers: headers?.toMap(), body: body);
    case HttpMethod.post:
      return _client.post(url, headers: headers?.toMap(), body: body);
    case HttpMethod.put:
      return _client.put(url, headers: headers?.toMap(), body: body);
  }
}

typedef Handler = FutureOr<Response> Function(Request request);

class ManualTestController extends TargetController {
  ManualTestController(
    this.path, {
    this.connectHandler = false,
    this.deleteHandler = false,
    this.getHandler = false,
    this.headHandler = false,
    this.optionsHandler = false,
    this.patchHandler = false,
    this.postHandler = false,
    this.putHandler = false,
    this.anyHandler = false,
    this.wsHandler = false,
  })  : anyHandlerCompleter = anyHandler ? Completer<Request>() : null,
        anyHandlerResponseCompleter = anyHandler ? Completer<Response>() : null,
        connectHandlerCompleter = connectHandler ? Completer<Request>() : null,
        connectHandlerResponseCompleter =
            connectHandler ? Completer<Response>() : null,
        deleteHandlerCompleter = deleteHandler ? Completer<Request>() : null,
        deleteHandlerResponseCompleter =
            deleteHandler ? Completer<Response>() : null,
        getHandlerCompleter = getHandler ? Completer<Request>() : null,
        getHandlerResponseCompleter = getHandler ? Completer<Response>() : null,
        headHandlerCompleter = headHandler ? Completer<Request>() : null,
        headHandlerResponseCompleter =
            headHandler ? Completer<Response>() : null,
        optionsHandlerCompleter = optionsHandler ? Completer<Request>() : null,
        optionsHandlerResponseCompleter =
            optionsHandler ? Completer<Response>() : null,
        patchHandlerCompleter = patchHandler ? Completer<Request>() : null,
        patchHandlerResponseCompleter =
            patchHandler ? Completer<Response>() : null,
        putHandlerCompleter = putHandler ? Completer<Request>() : null,
        putHandlerResponseCompleter = putHandler ? Completer<Response>() : null,
        postHandlerCompleter = postHandler ? Completer<Request>() : null,
        postHandlerResponseCompleter =
            postHandler ? Completer<Response>() : null,
        wsHandlerCompleter = wsHandler ? Completer<Stream<Object>>() : null;

  final String path;
  final Completer<Request>? anyHandlerCompleter;
  final Completer<Response>? anyHandlerResponseCompleter;
  final bool anyHandler;
  final Completer<Request>? connectHandlerCompleter;
  final Completer<Response>? connectHandlerResponseCompleter;
  final bool connectHandler;
  final Completer<Request>? deleteHandlerCompleter;
  final Completer<Response>? deleteHandlerResponseCompleter;
  final bool deleteHandler;
  final Completer<Request>? getHandlerCompleter;
  final Completer<Response>? getHandlerResponseCompleter;
  final bool getHandler;
  final Completer<Request>? headHandlerCompleter;
  final Completer<Response>? headHandlerResponseCompleter;
  final bool headHandler;
  final Completer<Request>? optionsHandlerCompleter;
  final Completer<Response>? optionsHandlerResponseCompleter;
  final bool optionsHandler;
  final Completer<Request>? patchHandlerCompleter;
  final Completer<Response>? patchHandlerResponseCompleter;
  final bool patchHandler;
  final Completer<Request>? postHandlerCompleter;
  final Completer<Response>? postHandlerResponseCompleter;
  final bool postHandler;
  final Completer<Request>? putHandlerCompleter;
  final Completer<Response>? putHandlerResponseCompleter;
  final bool putHandler;
  final Completer<Stream<Object>>? wsHandlerCompleter;
  final bool wsHandler;

  @override
  Future<Response> any(Request request) {
    if (anyHandlerCompleter != null) {
      anyHandlerCompleter!.complete(request);
    }
    return anyHandlerResponseCompleter?.future ??
        Future.value(
          Response.ok(null),
        );
  }

  @override
  Future<Response> connect(Request request) {
    if (connectHandlerCompleter != null) {
      connectHandlerCompleter!.complete(request);
    }
    return connectHandlerResponseCompleter?.future ??
        Future.value(
          Response.ok(null),
        );
  }

  @override
  Future<Response> delete(Request request) {
    if (deleteHandlerCompleter != null) {
      deleteHandlerCompleter!.complete(request);
    }
    return deleteHandlerResponseCompleter?.future ??
        Future.value(
          Response.ok(null),
        );
  }

  @override
  Future<Response> get(Request request) {
    if (getHandlerCompleter != null) {
      getHandlerCompleter!.complete(request);
    }
    return getHandlerResponseCompleter?.future ??
        Future.value(
          Response.ok(null),
        );
  }

  @override
  Future<Response> head(Request request) {
    if (headHandlerCompleter != null) {
      headHandlerCompleter!.complete(request);
    }
    return headHandlerResponseCompleter?.future ??
        Future.value(
          Response.ok(null),
        );
  }

  @override
  Future<Response> options(Request request) {
    if (optionsHandlerCompleter != null) {
      optionsHandlerCompleter!.complete(request);
    }
    return optionsHandlerResponseCompleter?.future ??
        Future.value(
          Response.ok(null),
        );
  }

  @override
  Future<Response> patch(Request request) {
    if (patchHandlerCompleter != null) {
      patchHandlerCompleter!.complete(request);
    }
    return patchHandlerResponseCompleter?.future ??
        Future.value(
          Response.ok(null),
        );
  }

  @override
  Future<Response> post(Request request) {
    if (postHandlerCompleter != null) {
      postHandlerCompleter!.complete(request);
    }
    return postHandlerResponseCompleter?.future ??
        Future.value(
          Response.ok(null),
        );
  }

  @override
  Future<Response> put(Request request) {
    if (putHandlerCompleter != null) {
      putHandlerCompleter!.complete(request);
    }
    return putHandlerResponseCompleter?.future ??
        Future.value(
          Response.ok(null),
        );
  }

  @override
  Stream<Object?> ws(Request request, WebSocketHandler websocket) async* {
    // if (wsHandlerCompleter != null) {
    //   wsHandlerCompleter!.complete(websocket);
    // }
    // return Stream.empty();
  }
}

class TestController extends TargetController {
  TestController(
    this.path, {
    this.middleware = const [],
    this.connectHandler,
    this.deleteHandler,
    this.getHandler,
    this.headHandler,
    this.optionsHandler,
    this.patchHandler,
    this.postHandler,
    this.putHandler,
    this.anyHandler,
    this.wsHandler,
  });

  @override
  final String path;

  @override
  final List<TargetMiddleware> middleware;

  final Handler? anyHandler;
  final Handler? connectHandler;
  final Handler? deleteHandler;
  final Handler? getHandler;
  final Handler? headHandler;
  final Handler? optionsHandler;
  final Handler? patchHandler;
  final Handler? postHandler;
  final Handler? putHandler;
  final Stream<Object?> Function(Request request, WebSocketHandler ws)?
      wsHandler;

  @override
  FutureOr<Response> any(Request request) {
    if (anyHandler == null) return super.any(request);
    return anyHandler!(request);
  }

  @override
  FutureOr<Response> connect(Request request) {
    if (connectHandler == null) return super.connect(request);
    return connectHandler!(request);
  }

  @override
  FutureOr<Response> delete(Request request) {
    if (deleteHandler == null) return super.delete(request);
    return deleteHandler!(request);
  }

  @override
  FutureOr<Response> get(Request request) {
    if (getHandler == null) return super.get(request);
    return getHandler!(request);
  }

  @override
  FutureOr<Response> head(Request request) {
    if (headHandler == null) return super.head(request);
    return headHandler!(request);
  }

  @override
  FutureOr<Response> options(Request request) {
    if (optionsHandler == null) return super.options(request);
    return optionsHandler!(request);
  }

  @override
  FutureOr<Response> patch(Request request) {
    if (patchHandler == null) return super.patch(request);
    return patchHandler!(request);
  }

  @override
  FutureOr<Response> post(Request request) {
    if (postHandler == null) return super.post(request);
    return postHandler!(request);
  }

  @override
  FutureOr<Response> put(Request request) {
    if (putHandler == null) return super.put(request);
    return putHandler!(request);
  }

  @override
  Stream<Object?> ws(Request request, WebSocketHandler websocket) {
    if (wsHandler == null) return super.ws(request, websocket);
    return wsHandler!(request, websocket);
  }
}

class TestProxyController extends TargetProxyController {
  TestProxyController({required this.path, required this.proxyHandler});

  final String path;
  final Future<Response> Function(Request request) proxyHandler;

  @override
  Future<Response> proxy(Request request) {
    return proxyHandler(request);
  }
}

class TestConfig extends TargetConfiguration {
  TestConfig({this.httpConfig, this.wsConfig});

  final HttpConfiguration? httpConfig;
  final WebSocketConfiguration? wsConfig;

  @override
  HttpConfiguration get http => httpConfig ?? HttpConfiguration();

  @override
  WebSocketConfiguration get ws => wsConfig ?? WebSocketConfiguration();
}

class TestMiddleware extends TargetMiddleware {
  TestMiddleware({this.handleRequest, this.handleResponse});

  final FutureOr<Response?> Function(Request request)? handleRequest;
  final FutureOr<Response> Function(Request request, Response response)?
      handleResponse;

  @override
  Future<Response?> requestHandler(Request request) async {
    return handleRequest?.call(request) ?? super.requestHandler(request);
  }

  @override
  Future<Response> responseHandler(Request request, Response response) async {
    return handleResponse?.call(request, response) ??
        super.responseHandler(request, response);
  }
}

class MockRequest extends Mock implements Request {
  @override
  HttpMethod get method {
    return super.noSuchMethod(
      Invocation.getter(#method),
      returnValue: HttpMethod.get,
      returnValueForMissingStub: HttpMethod.get,
    );
  }

  @override
  ImmutableHeaders get headers {
    return super.noSuchMethod(
      Invocation.getter(#headers),
      returnValue: ImmutableHeaders(),
      returnValueForMissingStub: ImmutableHeaders(),
    );
  }
}

class MockResponse extends Mock implements Response {
  @override
  Stream<List<int>> get body {
    return super.noSuchMethod(
      Invocation.getter(#body),
      returnValue: Stream<List<int>>.empty(),
      returnValueForMissingStub: Stream<List<int>>.empty(),
    );
  }

  @override
  int get status {
    return super.noSuchMethod(
      Invocation.getter(#status),
      returnValue: 200,
      returnValueForMissingStub: 200,
    );
  }

  @override
  ImmutableHeaders get headers {
    return super.noSuchMethod(
      Invocation.getter(#headers),
      returnValue: ImmutableHeaders(),
      returnValueForMissingStub: ImmutableHeaders(),
    );
  }
}

class FakeRequest extends Fake implements Request {
  FakeRequest({this.fakeHeaders, this.fakeMethod});

  final ImmutableHeaders? fakeHeaders;

  final HttpMethod? fakeMethod;

  @override
  HttpMethod get method {
    if (fakeMethod != null) return fakeMethod!;
    throw UnimplementedError();
  }

  @override
  ImmutableHeaders get headers {
    if (fakeHeaders != null) return fakeHeaders!;
    throw UnimplementedError();
  }
}

class RequestHandlerMock extends Mock {
  Future<Response?> call(Request? request) {
    return super.noSuchMethod(
      Invocation.method(#call, [request]),
      returnValue: Future.value(null),
      returnValueForMissingStub: Future.value(null),
    );
  }
}

class FakeResponse extends Fake implements Response {}

class ResponseHandlerMock extends Mock {
  FutureOr<Response> call(Request? request, Response? response) {
    return super.noSuchMethod(
      Invocation.method(#call, [request, response]),
      returnValue: Future.value(FakeResponse()),
      returnValueForMissingStub: Future.value(FakeResponse()),
    );
  }
}

class RequestChangeMock extends Mock {
  void call({Headers? headers, List<int>? body}) {
    return super.noSuchMethod(
      Invocation.method(#call, [
        {headers, body}
      ]),
      returnValue: '123',
      returnValueForMissingStub: '345',
    );
  }
}

class FakeTargetServer extends Fake implements TargetServer {}

class FakeShelfRequest extends Fake implements shelf.Request {
  FakeShelfRequest({this.requestHeaders});

  final Map<String, String>? requestHeaders;

  @override
  Map<String, String> get headers {
    if (requestHeaders != null) return requestHeaders!;
    throw UnimplementedError();
  }
}

class TestView extends View {
  TestView(String entry, this.content, [Map<String, dynamic>? data])
      : super(entry, data ?? {});

  Map<String, String> content;

  @override
  String getFileContent(String path) {
    return content[path]!;
  }
}
